INCLUDE(CMakeForceCompiler)

SET(CMAKE_SYSTEM_NAME Generic)
SET(CMAKE_SYSTEM_VERSION 1)
SET(CTR 1)
SET(CMAKE_HOST_SYSTEM_NAME Cygwin)
SET(CMAKE_HOST_SYSTEM_VERSION 1)
SET(ZLIB_FOUND 1)
SET(CMAKE_TRY_COMPILE_TARGET_TYPE STATIC_LIBRARY)
SET(CMAKE_CROSSCOMPILING TRUE)

IF(${CMAKE_BUILD_TYPE} MATCHES "Release")
	SET(CTR_BUILD_TYPE "release")
ELSEIF(${CMAKE_BUILD_TYPE} MATCHES "Debug")
	SET(CTR_BUILD_TYPE "noopt")
ELSE()
	SET(CTR_BUILD_TYPE "verbose")
ENDIF()

FIND_PROGRAM(CYGPATH cygpath)
IF(NOT CYGPATH)
	MESSAGE(FATAL_ERROR "cannot find cygpath")
ENDIF()
FUNCTION(nlib_cygwin_getcygpath_ _from _to)
	EXECUTE_PROCESS(COMMAND
		${CYGPATH} -u "${_from}"
		RESULT_VARIABLE cygpath_result
		ERROR_VARIABLE cygpath_err
		OUTPUT_VARIABLE _toraw
	)
	IF(cygpath_result OR cygpath_err)
		MESSAGE(FATAL_ERROR "${CYGPATH} failed:: ${cygpath_err}")
	ENDIF()
	STRING(REGEX REPLACE "[\r\n]" "" _toans "${_toraw}")
	GET_FILENAME_COMPONENT(_toans ${_toans} ABSOLUTE)
	SET(${_to} ${_toans} CACHE FILEPATH "" FORCE)
ENDFUNCTION()

FUNCTION(nlib_cygwin_getwinpath_ _from _to)
	GET_FILENAME_COMPONENT(_from ${_from} ABSOLUTE)
	EXECUTE_PROCESS(COMMAND
		${CYGPATH} -m "${_from}"
		RESULT_VARIABLE cygpath_result
		ERROR_VARIABLE cygpath_err
		OUTPUT_VARIABLE _toraw
	)
	IF(cygpath_result OR cygpath_err)
		MESSAGE(FATAL_ERROR "${CYGPATH} failed:: ${cygpath_err}")
	ENDIF()
	STRING(REGEX REPLACE "[\r\n]" "" _toans "${_toraw}")
	SET(${_to} ${_toans} CACHE FILEPATH "" FORCE)
ENDFUNCTION()

FUNCTION(join VALUES GLUE OUTPUT)
  STRING(REPLACE ";" "${GLUE}" _TMP_STR "${VALUES}")
  SET(${OUTPUT} "${_TMP_STR}" PARENT_SCOPE)
ENDFUNCTION()

nlib_cygwin_getcygpath_($ENV{ARMCC41BIN}/../ CYG_ARMCC41_ROOT)
nlib_cygwin_getwinpath_(${CYG_ARMCC41_ROOT} WIN_ARMCC41_ROOT)
nlib_cygwin_getcygpath_($ENV{CTRSDK_ROOT} CYG_CTRSDK_ROOT)
nlib_cygwin_getwinpath_(${CYG_CTRSDK_ROOT} WIN_CTRSDK_ROOT)
MESSAGE(STATUS "WIN_CTRSDK_ROOT=${WIN_CTRSDK_ROOT}")
MESSAGE(STATUS "CYG_CTRSDK_ROOT=${CYG_CTRSDK_ROOT}")
MESSAGE(STATUS "WIN_ARMCC41_ROOT=${WIN_ARMCC41_ROOT}")
MESSAGE(STATUS "CYG_ARMCC41_ROOT=${CYG_ARMCC41_ROOT}")

# specify the cross compiler
#  set(CMAKE_C_COMPILE_OBJECT
#    "<CMAKE_C_COMPILER> <DEFINES> <FLAGS> -o <OBJECT> -c <SOURCE>")
#  set(CMAKE_CXX_COMPILE_OBJECT
#    "<CMAKE_CXX_COMPILER>  <DEFINES> <FLAGS> -o <OBJECT> -c <SOURCE>")
SET(CMAKE_C_COMPILER "${CYG_ARMCC41_ROOT}/bin/armcc.exe")
#CMAKE_FORCE_C_COMPILER(${CYG_ARMCC41_ROOT}/bin/armcc.exe ARMCC)
set(CMAKE_C_COMPILER_ID_RUN TRUE)
set(CMAKE_C_COMPILER_ID GHS)
set(CMAKE_C_COMPILER_FORCED TRUE)
SET(CMAKE_CXX_COMPILER "${CYG_ARMCC41_ROOT}/bin/armcc.exe")
#CMAKE_FORCE_CXX_COMPILER(${CYG_ARMCC41_ROOT}/bin/armcc.exe ARMCC)
set(CMAKE_CXX_COMPILER_ID_RUN TRUE)
set(CMAKE_CXX_COMPILER_ID GHS)
set(CMAKE_CXX_COMPILER_FORCED TRUE)
SET(CMAKE_AR "${CYG_ARMCC41_ROOT}/bin/armar.exe" CACHE FILEPATH "Archiver" FORCE)
SET(CMAKE_DEBUG_POSTFIX ".fast" CACHE FILEPATH "" FORCE)
SET(CMAKE_RELEASE_POSTFIX ".fast" CACHE FILEPATH "" FORCE)
SET(CMAKE_RELWITHDEBINFO_POSTFIX ".fast" CACHE FILEPATH "" FORCE)

SET(tmp
	"echo <DEFINES> <INCLUDES> <FLAGS> -o <OBJECT> -c <SOURCE>"
	"awk -f \"${CMAKE_CURRENT_LIST_DIR}/convert_include_path.awk\""
	"xargs")
join("${tmp}" " | " tmp)
SET(CMAKE_C_COMPILE_OBJECT "${tmp} <CMAKE_C_COMPILER>")
SET(CMAKE_CXX_COMPILE_OBJECT "${tmp} <CMAKE_CXX_COMPILER>")

SET(CMAKE_CXX_ARCHIVE_CREATE "<CMAKE_AR> -rcuT <TARGET> <LINK_FLAGS> <OBJECTS>")
SET(CMAKE_C_ARCHIVE_CREATE "<CMAKE_AR> -rcuT <TARGET> <LINK_FLAGS> <OBJECTS>")

# set CTR_SDK_LIBS if you have to add other libs
# workaround for cmake 3.6.2
#LINK_DIRECTORIES(${CYG_CTRSDK_ROOT}/libraries/CTR-TS.Process.MPCore/${CTR_BUILD_TYPE})

MESSAGE(STATUS "include directories: ${CYG_CTRSDK_ROOT}/include/")
INCLUDE_DIRECTORIES(${CYG_CTRSDK_ROOT}/include/)

SET(tmp
	--cpp --debug_info=line_inlining_extensions --force_new_nothrow --fpmode=fast
	--no_exceptions --no_rtti --remarks --vfe --diag_suppress 186,340,401,1256,1297
	--diag_suppress 1568,1764,1786,1788,2523,2819 --diag_suppress 96,1794,1801,2442
	--diag_error 68,88,174,188,223 --diag_warning 177,193,228,550
	--diag_suppress 3017 --diag_suppress=optimizations --diag_suppress=1301,2530,237
	--apcs /interwork --data_reorder --debug --diag_style=arm --dwarf3
	--no_debug_macros --remove_unneeded_entities --split_sections --dollar
	--littleend --wchar --wchar16 --multibyte_chars --signed_chars --cpu MPCore --arm)
JOIN("${tmp}" " " tmp)
SET(CMAKE_CXX_FLAGS_INIT ${tmp})
SET(tmp
	--debug_info=line_inlining_extensions --force_new_nothrow --fpmode=fast
	--no_exceptions --no_rtti --remarks --vfe --diag_suppress 186,340,401,1256,1297
	--diag_suppress 1568,1764,1786,1788,2523,2819 --diag_suppress 96,1794,1801,2442
	--diag_error 68,88,174,188,223 --diag_warning 177,193,228,550
	--diag_suppress 3017 --diag_suppress=optimizations --diag_suppress=1301,2530,237
	--apcs /interwork --data_reorder --debug --diag_style=arm --dwarf3
	--no_debug_macros --remove_unneeded_entities --split_sections --dollar
	--littleend --wchar --wchar16 --multibyte_chars --signed_chars --cpu MPCore --arm)
JOIN("${tmp}" " " tmp)

SET(tmpdefs "-DNN_COMPILER_RVCT -DNN_COMPILER_RVCT_VERSION_MAJOR=4 -DNN_PROCESSOR_ARM")
SET(tmpdefs "${tmpdefs} -DNN_PROCESSOR_ARM11MPCORE -DNN_PROCESSOR_ARM_V6 -DNN_PROCESSOR_ARM_VFP_V2")
SET(tmpdefs "${tmpdefs} -DNN_HARDWARE_CTR -DNN_EFFORT_FAST -DNN_PLATFORM_CTR -DNN_HARDWARE_CTR_TS")
SET(tmpdefs "${tmpdefs} -DNN_SYSTEM_PROCESS -DNN_COMPILER_OPTION_ARM -DNN_DEBUGGER_KMC_PARTNER")

SET(CMAKE_C_FLAGS "${tmp} ${tmpdefs}" CACHE STRING "" FORCE)
SET(CMAKE_CXX_FLAGS "${tmp} ${tmpdefs}" CACHE STRING "" FORCE)

SET(CMAKE_CXX_FLAGS_DEBUG "--gnu -Otime -O0 --retain=calls -DNN_SWITCH_ENABLE_HOST_IO=1 -DNN_BUILD_VERBOSE -DNN_BUILD_NOOPT" CACHE STRING "" FORCE)
SET(CMAKE_CXX_FLAGS_RELEASE	"--gnu -Otime -O3 -DNN_SWITCH_DISABLE_DEBUG_PRINT=1 -DNN_SWITCH_DISABLE_DEBUG_PRINT_FOR_SDK=1 -DNN_SWITCH_DISABLE_ASSERT_WARNING=1 -DNN_SWITCH_DISABLE_ASSERT_WARNING_FOR_SDK=1" CACHE STRING "" FORCE)
SET(CMAKE_CXX_FLAGS_RELWITHDEBINFO "--gnu -Otime -O3 -DNN_SWITCH_ENABLE_HOST_IO=1 -DNN_BUILD_VERBOSE" CACHE STRING "" FORCE)

SET(CMAKE_C_FLAGS_DEBUG	"--gnu -Otime -O0 --retain=calls -DNN_SWITCH_ENABLE_HOST_IO=1 -DNN_BUILD_VERBOSE -DNN_BUILD_NOOPT" CACHE STRING "" FORCE)
SET(CMAKE_C_FLAGS_RELEASE "--gnu -Otime -O3 -DNN_SWITCH_DISABLE_DEBUG_PRINT=1 -DNN_SWITCH_DISABLE_DEBUG_PRINT_FOR_SDK=1 -DNN_SWITCH_DISABLE_ASSERT_WARNING=1 -DNN_SWITCH_DISABLE_ASSERT_WARNING_FOR_SDK=1" CACHE STRING "" FORCE)
SET(CMAKE_C_FLAGS_RELWITHDEBINFO "--gnu -Otime -O3 -DNN_SWITCH_ENABLE_HOST_IO=1 -DNN_BUILD_VERBOSE" CACHE STRING "" FORCE)

SET(tmp
	architecture exceptions inline inputs libraries merge sizes stack
	summarysizes summarystack tailreorder totals unused unusedsymbols
	veneerpools veneers veneercallers visibility)
join("${tmp}" "," _info)

SET(tmp
	--exceptions_tables=nocreate --no_exceptions --scanlib --vfemode=force
	--callgraph --callgraph_output=text --verbose --map --symbols --xref --unmangled
	--info=${_info}
	--diag_suppress L6314W,L6329W --debug --diag_style=arm
	--diag_suppress=1301,2530,237 --largeregions --merge --no_eager_load_debug
	--pad 0xFF --remove --strict --entry=__ctr_start --keep=nnMain
	--startup=__ctr_start --datacompressor off --bestdebug --cpu MPCore)

join("${tmp}" " " tmp)
SET(CMAKE_CXX_LINK_FLAGS ${tmp})
SET(CMAKE_C_LINK_FLAGS ${tmp})

SET(echoarg
	"-o <TARGET_BASE>.axf"
	"--list=<TARGET_BASE>.map"
	"--callgraph_file=<TARGET_BASE>.txt"
	"--userlibpath=${WIN_CTRSDK_ROOT}/libraries/CTR-TS.Process.MPCore/${CTR_BUILD_TYPE} <LINK_LIBRARIES>"
	"--scatter=${WIN_CTRSDK_ROOT}/resources/specfiles/linker/CTR.Process.MPCore.ldscript"
	"<OBJECTS>"
	"${WIN_CTRSDK_ROOT}/libraries/CTR-TS.Process.MPCore/${CTR_BUILD_TYPE}/crt0.fast.o")
join("${echoarg}" " " echoarg)

SET(filter
	"awk -v RS=' ' '$0!~/-L[^ ]/{print}/-L[^ ]/{cmd=\\\"cygpath -m \\\"substr($0, 3)\;cmd|getline\;print \\\"-L\\\" $0}'|awk -v ORS=' ' '{print}' | sed -r -e 's/ -l\\([^ ]*\\)/ --library=\\\\1.fast/g' -e 's/ -L\\([^ ]*\\)/\\\\1,/g'")

SET(makerom "\"${CYG_CTRSDK_ROOT}/tools/CommandLineTools/ctr_makerom32.exe\"")
SET(makebin
	"\"${CYG_ARMCC41_ROOT}/bin/armlink.exe\" --via <OBJECT_DIR>/linktmp.sh"
	"\"${CYG_ARMCC41_ROOT}/bin/fromelf.exe\" --32x1 -cd -o <TARGET_BASE>.dasm <TARGET_BASE>.axf"
	"sh -c \"cat <OBJECT_DIR>/makeromopt.txt | xargs ${makerom} -o <TARGET_BASE>.cci <TARGET_BASE>.axf")

# -L xxx -L yyy -> --userlibpath=xxx,yyy
# -l nx_misc -> --library=nx_misc.fast
SET(CMAKE_CXX_LINK_EXECUTABLE
	"sh -c \"echo <CMAKE_CXX_LINK_FLAGS> ${echoarg} | ${filter} > <OBJECT_DIR>/linktmp.sh\""
	"${makebin}"
)
SET(CMAKE_C_LINK_EXECUTABLE
	"sh -c \"echo <CMAKE_C_LINK_FLAGS> ${echoarg} | ${filter} > <OBJECT_DIR>/linktmp.sh\""
	"${makebin}"
)

#<CMAKE_CXX_COMPILER>  <FLAGS> <CMAKE_CXX_LINK_FLAGS> <LINK_FLAGS> <OBJECTS>  -o <TARGET> <LINK_LIBRARIES>")
#  set(CMAKE_C_LINK_EXECUTABLE
#   "<CMAKE_C_COMPILER> <FLAGS> <CMAKE_C_LINK_FLAGS> <LINK_FLAGS> <OBJECTS>  -o <TARGET> <LINK_LIBRARIES>")

SET(CMAKE_SIZEOF_VOID_P 4 CACHE INTERNAL "")
SET(CMAKE_EXECUTABLE_SUFFIX ".cci" CACHE INTERNAL "")

# search for programs in the build host directories
SET(CMAKE_FIND_ROOT_PATH_MODE_PROGRAM NEVER)
# for libraries and headers in the target directories
SET(CMAKE_FIND_ROOT_PATH_MODE_LIBRARY ONLY)
SET(CMAKE_FIND_ROOT_PATH_MODE_INCLUDE ONLY)
SET(CMAKE_FIND_ROOT_PATH_MODE_PACKAGE ONLY)
