﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HASH_H_
#define INCLUDE_NN_NLIB_HASH_H_

#include <functional>
#include "nn/nlib/Config.h"

#if defined(NLIB_CXX11_TEMPLATE_ALIAS)

template<class T> using Hash = std::hash<T>;

#define NLIB_DEFINE_STD_HASH_BEGIN1(ns1) namespace std { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN2(ns1, ns2) namespace std { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN3(ns1, ns2, ns3) namespace std { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN4(ns1, ns2, ns3, ns4) namespace std { // NOLINT
#define NLIB_DEFINE_STD_HASH_END1(ns1) } // NOLINT
#define NLIB_DEFINE_STD_HASH_END2(ns1, ns2) } // NOLINT
#define NLIB_DEFINE_STD_HASH_END3(ns1, ns2, ns3) } // NOLINT
#define NLIB_DEFINE_STD_HASH_END4(ns1, ns2, ns3, ns4) } // NOLINT
#define NLIB_DEFINE_STD_HASH_T(tp) \
template<> struct hash<tp> { \
    size_t operator()(const tp& val) const NLIB_NOEXCEPT { return val.GetHash(); } \
};
#define NLIB_DEFINE_STD_HASH_T2(targ1, targ2, tp) \
template<class targ1, class targ2> struct hash<tp<targ1, targ2> > { \
    size_t operator()(const tp<targ1, targ2>& val) const NLIB_NOEXCEPT { return val.GetHash(); } \
};

#else
NLIB_NAMESPACE_BEGIN

namespace detail {
inline size_t Times33(const unsigned char* p, size_t n) NLIB_NOEXCEPT {
    size_t hash = 0;
    for (; n != 0; --n, ++p) { hash = hash * 33 + *p; }
    return hash;
}
}  // namespace detail

template<class T>
struct SimpleHash : public std::unary_function<T, size_t> {
    size_t operator()(const T& val) const NLIB_NOEXCEPT {
        return detail::Times33(
            reinterpret_cast<const unsigned char*>(&val), sizeof(val));
    }
};

template<class T> struct Hash : public SimpleHash<T> {};

template<> struct Hash<bool> : public SimpleHash<bool> {};
template<> struct Hash<char> : public SimpleHash<char> {};
template<> struct Hash<signed char> : public SimpleHash<signed char> {};
template<> struct Hash<unsigned char> : public SimpleHash<unsigned char> {};
template<> struct Hash<wchar_t> : public SimpleHash<wchar_t> {};
template<> struct Hash<short> : public SimpleHash<short> {}; // NOLINT
template<> struct Hash<unsigned short> : public SimpleHash<unsigned short> {}; // NOLINT
template<> struct Hash<int> : public SimpleHash<int> {};
template<> struct Hash<unsigned int> : public SimpleHash<unsigned int> {};
template<> struct Hash<long> : public SimpleHash<long> {}; // NOLINT
template<> struct Hash<unsigned long> : public SimpleHash<unsigned long> {}; // NOLINT
template<> struct Hash<long long> : public SimpleHash<long long> {}; // NOLINT
template<> struct Hash<unsigned long long> : public SimpleHash<unsigned long long> {}; // NOLINT
template<> struct Hash<float> : public SimpleHash<float> {};
template<> struct Hash<double> : public SimpleHash<double> {};
template<class T> struct Hash<T*> : public SimpleHash<T*> {};

// NOTE:
// This file defines the subset of C++11 std::hash.
// No definitions for std::string, std::wstring for now
#define NLIB_DEFINE_STD_HASH_BEGIN1(ns1) namespace ns1 { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN2(ns1, ns2) namespace ns1 { namespace ns2 { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN3(ns1, ns2, ns3) \
    namespace ns1 { namespace ns2 { namespace ns3 { // NOLINT
#define NLIB_DEFINE_STD_HASH_BEGIN4(ns1, ns2, ns3, ns4) \
    namespace ns1 { namespace ns2 { namespace ns3 { namespace ns4 { // NOLINT
#define NLIB_DEFINE_STD_HASH_END1(ns1) } // NOLINT
#define NLIB_DEFINE_STD_HASH_END2(ns1, ns2) }} // NOLINT
#define NLIB_DEFINE_STD_HASH_END3(ns1, ns2, ns3) }}} // NOLINT
#define NLIB_DEFINE_STD_HASH_END4(ns1, ns2, ns3, ns4) }}}} // NOLINT

#define NLIB_DEFINE_STD_HASH_T(tp) \
template<> struct Hash<tp> { \
    size_t operator()(const tp& val) const NLIB_NOEXCEPT { return val.GetHash(); } \
};
#define NLIB_DEFINE_STD_HASH_T2(targ1, targ2, tp) \
template<class targ1, class targ2> struct Hash<tp<targ1, targ2> > { \
    size_t operator()(const tp<targ1, targ2>& val) const NLIB_NOEXCEPT { return val.GetHash(); } \
};


NLIB_NAMESPACE_END
#endif



#endif  // INCLUDE_NN_NLIB_HASH_H_
