﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_PLATFORM_H_
#define INCLUDE_NN_NLIB_PLATFORM_H_

#ifdef __ILP64__
# error Sorry, ILP64 not supported
#endif

// Note that the C preprocessor # operator must be able to use.
#define NLIB_VERSION_YEAR 2016
#define NLIB_VERSION_YEAR_SHORT 16
#define NLIB_VERSION_DATE 1222
#define NLIB_VERSION 20161222

#ifdef __cplusplus
# ifndef __STDC_FORMAT_MACROS
#  ifdef _MSC_VER
#   pragma message(__FILE__ ": __STDC_FORMAT_MACROS not defined, compile may fail")
#  elif !defined(NN_PLATFORM_CTR) && !defined(CAFE)
#   warning __STDC_FORMAT_MACROS not defined, compile may fail
#  endif
#  define __STDC_FORMAT_MACROS
# endif
#endif
#ifndef __STDC_WANT_LIB_EXT1__
# define __STDC_WANT_LIB_EXT1__ 1
#endif

#ifndef NLIB_UNUSED
# define NLIB_UNUSED(x) (void)(x)
#endif

#ifndef __has_builtin
# define __has_builtin(x) 0
#endif

#ifndef __has_feature
# define __has_feature(x) 0
#endif

#ifndef __has_cpp_attribute
# define __has_cpp_attribute(x) 0
#endif

#ifndef __has_attribute
# define __has_attribute(x) 0
#endif

#ifndef __has_declspec_attribute
# define __has_declspec_attribute(x) 0
#endif

#ifndef __has_include
# define __has_include(x) 0
#endif

#if __has_feature(thread_sanitizer)
# define NLIB_NO_TSAN __attribute__((no_sanitize("thread")))
#else
# define NLIB_NO_TSAN
#endif

#if __has_feature(address_sanitizer)
# define NLIB_NO_ASAN __attribute__((no_sanitize("address")))
#else
# define NLIB_NO_ASAN
#endif

#if __has_feature(memory_sanitizer)
# define NLIB_NO_MSAN __attribute__((no_sanitize("memory")))
#else
# define NLIB_NO_MSAN
#endif

//
// thread safety analysis
// http://clang.llvm.org/docs/ThreadSafetyAnalysis.html
//
#if !defined(NLIB_THREAD_AA_) && defined(__clang__) && \
    (!defined(SWIG)) && __has_attribute(capability)
#define NLIB_THREAD_AA_(x) __attribute__((x))
#else
#ifdef NLIB_THREAD_AA_
#undef NLIB_THREAD_AA_
#endif
#define NLIB_THREAD_AA_(x)
#endif

#define NLIB_CAPABILITY(x) NLIB_THREAD_AA_(capability(x))
#define NLIB_SCOPED_CAPABILITY NLIB_THREAD_AA_(scoped_lockable)
#define NLIB_GUARDED_BY(x) NLIB_THREAD_AA_(guarded_by(x))
#define NLIB_PT_GUARDED_BY(x) NLIB_THREAD_AA_(pt_guarded_by(x))
#define NLIB_ACQUIRED_BEFORE(...) NLIB_THREAD_AA_(acquired_before(__VA_ARGS__))
#define NLIB_ACQUIRED_AFTER(...) NLIB_THREAD_AA_(acquired_after(__VA_ARGS__))
#define NLIB_REQUIRES(...) NLIB_THREAD_AA_(requires_capability(__VA_ARGS__))
#define NLIB_REQUIRES_SHARED(...) NLIB_THREAD_AA_(requires_shared_capability(__VA_ARGS__))
#define NLIB_ACQUIRE(...) NLIB_THREAD_AA_(acquire_capability(__VA_ARGS__))
#define NLIB_ACQUIRE_SHARED(...) NLIB_THREAD_AA_(acquire_shared_capability(__VA_ARGS__))
#define NLIB_RELEASE(...) NLIB_THREAD_AA_(release_capability(__VA_ARGS__))
#define NLIB_RELEASE_SHARED(...) NLIB_THREAD_AA_(release_shared_capability(__VA_ARGS__))
#define NLIB_TRY_ACQUIRE(...) NLIB_THREAD_AA_(try_acquire_capability(__VA_ARGS__))
#define NLIB_TRY_ACQUIRE_SHARED(...) NLIB_THREAD_AA_(try_acquire_shared_capability(__VA_ARGS__))
#define NLIB_EXCLUDES(...) NLIB_THREAD_AA_(locks_excluded(__VA_ARGS__))
#define NLIB_ASSERT_CAPABILITY(x) NLIB_THREAD_AA_(assert_capability(x))
#define NLIB_ASSERT_SHARED_CAPABILITY(x) NLIB_THREAD_AA_(assert_shared_capability(x))
#define NLIB_RETURN_CAPABILITY(x) NLIB_THREAD_AA_(lock_returned(x))
#define NLIB_NO_THREAD_SAFETY_ANALYSIS NLIB_THREAD_AA_(no_thread_safety_analysis)

#include <stddef.h>  // for size_t
#include <stdio.h>   // for SEEK_CUR, SEEK_END, SEEK_SET
#include <stdarg.h>  // for va_list
#include <string.h>  // for memcpy, memmove, memset
#include <time.h>    // for struct timespec

#if defined(NLIB_RENAME_CAPI)
#ifndef NLIB_CFUNC_PREFIX
# define NLIB_CFUNC_PREFIX mypfx
#endif
#ifndef NLIB_CAPI
# define NLIB_CAPI__(p, q) p ## q
# define NLIB_CAPI_(p, q) NLIB_CAPI__(p, q)
# define NLIB_CAPI(x) NLIB_CAPI_(NLIB_CFUNC_PREFIX, x)
#endif

#define nlib_error_string NLIB_CAPI(nlib_error_string)
#define nlib_get_native_last_error NLIB_CAPI(nlib_get_native_last_error)
#define nlib_getversion NLIB_CAPI(nlib_getversion)
#define nlib_compiler_version NLIB_CAPI(nlib_compiler_version)
#define nlib_crc32 NLIB_CAPI(nlib_crc32)
#define nlib_crc32c NLIB_CAPI(nlib_crc32c)
#define nlib_epochtime NLIB_CAPI(nlib_epochtime)
#define nlib_ticktime NLIB_CAPI(nlib_ticktime)
#define nlib_sleep NLIB_CAPI(nlib_sleep)
#define nlib_timer_create NLIB_CAPI(nlib_timer_create)
#define nlib_timer_settime NLIB_CAPI(nlib_timer_settime)
#define nlib_timer_gettime NLIB_CAPI(nlib_timer_gettime)
#define nlib_timer_delete NLIB_CAPI(nlib_timer_delete)
#define nlib_gen_random NLIB_CAPI(nlib_gen_random)
#define nlib_mempagesize NLIB_CAPI(nlib_mempagesize)
#define nlib_virtual_alloc NLIB_CAPI(nlib_virtual_alloc)
#define nlib_virtual_free NLIB_CAPI(nlib_virtual_free)
#define nlib_physical_alloc NLIB_CAPI(nlib_physical_alloc)
#define nlib_physical_free NLIB_CAPI(nlib_physical_free)
#define nlib_mlock NLIB_CAPI(nlib_mlock)
#define nlib_munlock NLIB_CAPI(nlib_munlock)
#define nlib_tls_alloc NLIB_CAPI(nlib_tls_alloc)
#define nlib_tls_free NLIB_CAPI(nlib_tls_free)
#define nlib_tls_setvalue NLIB_CAPI(nlib_tls_setvalue)
#define nlib_tls_getvalue NLIB_CAPI(nlib_tls_getvalue)
#define nlib_mutex_init NLIB_CAPI(nlib_mutex_init)
#define nlib_mutex_recursive_init NLIB_CAPI(nlib_mutex_recursive_init)
#define nlib_mutex_recursive_timed_init NLIB_CAPI(nlib_mutex_recursive_timed_init)
#define nlib_mutex_lock NLIB_CAPI(nlib_mutex_lock)
#define nlib_mutex_trylock NLIB_CAPI(nlib_mutex_trylock)
#define nlib_mutex_trylock_for NLIB_CAPI(nlib_mutex_trylock_for)
#define nlib_mutex_unlock NLIB_CAPI(nlib_mutex_unlock)
#define nlib_mutex_destroy NLIB_CAPI(nlib_mutex_destroy)
#define nlib_semaphore_init NLIB_CAPI(nlib_semaphore_init)
#define nlib_semaphore_wait NLIB_CAPI(nlib_semaphore_wait)
#define nlib_semaphore_trywait NLIB_CAPI(nlib_semaphore_trywait)
#define nlib_semaphore_trywait_for NLIB_CAPI(nlib_semaphore_trywait_for)
#define nlib_semaphore_post NLIB_CAPI(nlib_semaphore_post)
#define nlib_semaphore_post_ex NLIB_CAPI(nlib_semaphore_post_ex)
#define nlib_semaphore_destroy NLIB_CAPI(nlib_semaphore_destroy)
#define nlib_cond_init NLIB_CAPI(nlib_cond_init)
#define nlib_cond_signal NLIB_CAPI(nlib_cond_signal)
#define nlib_cond_broadcast NLIB_CAPI(nlib_cond_broadcast)
#define nlib_cond_wait NLIB_CAPI(nlib_cond_wait)
#define nlib_cond_wait_for NLIB_CAPI(nlib_cond_wait_for)
#define nlib_cond_wait_until NLIB_CAPI(nlib_cond_wait_until)
#define nlib_cond_destroy NLIB_CAPI(nlib_cond_destroy)
#define nlib_rwlock_init NLIB_CAPI(nlib_rwlock_init)
#define nlib_rwlock_destroy NLIB_CAPI(nlib_rwlock_destroy)
#define nlib_rwlock_rdlock NLIB_CAPI(nlib_rwlock_rdlock)
#define nlib_rwlock_tryrdlock NLIB_CAPI(nlib_rwlock_tryrdlock)
#define nlib_rwlock_tryrdlock_for NLIB_CAPI(nlib_rwlock_tryrdlock_for)
#define nlib_rwlock_tryrdlock_until NLIB_CAPI(nlib_rwlock_tryrdlock_until)
#define nlib_rwlock_rdunlock NLIB_CAPI(nlib_rwlock_rdunlock)
#define nlib_rwlock_wrlock NLIB_CAPI(nlib_rwlock_wrlock)
#define nlib_rwlock_trywrlock NLIB_CAPI(nlib_rwlock_trywrlock)
#define nlib_rwlock_trywrlock_for NLIB_CAPI(nlib_rwlock_trywrlock_for)
#define nlib_rwlock_trywrlock_until NLIB_CAPI(nlib_rwlock_trywrlock_until)
#define nlib_rwlock_wrunlock NLIB_CAPI(nlib_rwlock_wrunlock)
#define nlib_condrwlock_init NLIB_CAPI(nlib_condrwlock_init)
#define nlib_condrwlock_destroy NLIB_CAPI(nlib_condrwlock_destroy)
#define nlib_condrwlock_signal NLIB_CAPI(nlib_condrwlock_signal)
#define nlib_condrwlock_broadcast NLIB_CAPI(nlib_condrwlock_broadcast)
#define nlib_condrwlock_wait NLIB_CAPI(nlib_condrwlock_wait)
#define nlib_condrwlock_wait_for NLIB_CAPI(nlib_condrwlock_wait_for)
#define nlib_condrwlock_wait_until NLIB_CAPI(nlib_condrwlock_wait_until)
#define nlib_barrier_init NLIB_CAPI(nlib_barrier_init)
#define nlib_barrier_destroy NLIB_CAPI(nlib_barrier_destroy)
#define nlib_barrier_wait NLIB_CAPI(nlib_barrier_wait)
#define nlib_once NLIB_CAPI(nlib_once)
#define nlib_tryonce NLIB_CAPI(nlib_tryonce)
#define nlib_mq_open NLIB_CAPI(nlib_mq_open)
#define nlib_mq_getattr NLIB_CAPI(nlib_mq_getattr)
#define nlib_mq_close NLIB_CAPI(nlib_mq_close)
#define nlib_mq_readonly NLIB_CAPI(nlib_mq_readonly)
#define nlib_mq_send NLIB_CAPI(nlib_mq_send)
#define nlib_mq_send_until NLIB_CAPI(nlib_mq_send_until)
#define nlib_mq_receive NLIB_CAPI(nlib_mq_receive)
#define nlib_mq_receive_until NLIB_CAPI(nlib_mq_receive_until)
#define nlib_mq_drop NLIB_CAPI(nlib_mq_drop)
#define nlib_yield NLIB_CAPI(nlib_yield)
#define nlib_thread_create NLIB_CAPI(nlib_thread_create)
#define nlib_thread_join NLIB_CAPI(nlib_thread_join)
#define nlib_thread_detach NLIB_CAPI(nlib_thread_detach)
#define nlib_thread_self NLIB_CAPI(nlib_thread_self)
#define nlib_thread_getconcurrency NLIB_CAPI(nlib_thread_getconcurrency)
#define nlib_thread_getid NLIB_CAPI(nlib_thread_getid)
#define nlib_thread_equal NLIB_CAPI(nlib_thread_equal)
#define nlib_thread_getcpu NLIB_CAPI(nlib_thread_getcpu)
#define nlib_thread_setaffinity NLIB_CAPI(nlib_thread_setaffinity)
#define nlib_thread_setname NLIB_CAPI(nlib_thread_setname)
#define nlib_thread_getname NLIB_CAPI(nlib_thread_getname)
#define nlib_thread_attr_init NLIB_CAPI(nlib_thread_attr_init)
#define nlib_thread_attr_setint NLIB_CAPI(nlib_thread_attr_setint)
#define nlib_thread_attr_getint NLIB_CAPI(nlib_thread_attr_getint)
#define nlib_thread_attr_setptr NLIB_CAPI(nlib_thread_attr_setptr)
#define nlib_thread_attr_getptr NLIB_CAPI(nlib_thread_attr_getptr)
#define nlib_thread_attr_setstack NLIB_CAPI(nlib_thread_attr_setstack)
#define nlib_thread_attr_getstack NLIB_CAPI(nlib_thread_attr_getstack)
#define nlib_thread_attr_destroy NLIB_CAPI(nlib_thread_attr_destroy)
#define nlib_thread_getpriority NLIB_CAPI(nlib_thread_getpriority)
#define nlib_thread_setpriority NLIB_CAPI(nlib_thread_setpriority)
#define nlib_thread_priority_min NLIB_CAPI(nlib_thread_priority_min)
#define nlib_thread_priority_max NLIB_CAPI(nlib_thread_priority_max)
#define nlib_thread_priority_default NLIB_CAPI(nlib_thread_priority_default)
#define nlib_thread_exit NLIB_CAPI(nlib_thread_exit)
#define nlib_thread_exit_cpp NLIB_CAPI(nlib_thread_exit_cpp)
#define nlib_thread_cleanup_push_ NLIB_CAPI(nlib_thread_cleanup_push_)
#define nlib_thread_cleanup_pop_ NLIB_CAPI(nlib_thread_cleanup_pop_)
#define nlib_write_stdout NLIB_CAPI(nlib_write_stdout)
#define nlib_write_stderr NLIB_CAPI(nlib_write_stderr)
#define nlib_debug_break NLIB_CAPI(nlib_debug_break)
#define nlib_debug_backtrace NLIB_CAPI(nlib_debug_backtrace)
#define nlib_debug_backtrace_gettext NLIB_CAPI(nlib_debug_backtrace_gettext)
#define nlib_getenv NLIB_CAPI(nlib_getenv)
#define nlib_log_print NLIB_CAPI(nlib_log_print)
#define nlib_log_vprint NLIB_CAPI(nlib_log_vprint)
#define nlib_log_attr_setint NLIB_CAPI(nlib_log_attr_setint)
#define nlib_fd_open NLIB_CAPI(nlib_fd_open)
#define nlib_fd_close NLIB_CAPI(nlib_fd_close)
#define nlib_fd_read NLIB_CAPI(nlib_fd_read)
#define nlib_fd_write NLIB_CAPI(nlib_fd_write)
#define nlib_fd_seek NLIB_CAPI(nlib_fd_seek)
#define nlib_fd_pread NLIB_CAPI(nlib_fd_pread)
#define nlib_fd_pwrite NLIB_CAPI(nlib_fd_pwrite)
#define nlib_fd_truncate NLIB_CAPI(nlib_fd_truncate)
#define nlib_fd_getsize NLIB_CAPI(nlib_fd_getsize)
#define nlib_fd_flush NLIB_CAPI(nlib_fd_flush)
#define nlib_fd_sync NLIB_CAPI(nlib_fd_sync)
#define nlib_fd_native_handle NLIB_CAPI(nlib_fd_native_handle)
#define nlib_fd_readv NLIB_CAPI(nlib_fd_readv)
#define nlib_fd_writev NLIB_CAPI(nlib_fd_writev)
#define nlib_fd_preadv NLIB_CAPI(nlib_fd_preadv)
#define nlib_fd_pwritev NLIB_CAPI(nlib_fd_pwritev)
#define nlib_unlink NLIB_CAPI(nlib_unlink)
#define nlib_remove NLIB_CAPI(nlib_remove)
#define nlib_mkdir NLIB_CAPI(nlib_mkdir)
#define nlib_rmdir NLIB_CAPI(nlib_rmdir)
#define nlib_rename NLIB_CAPI(nlib_rename)
#define nlib_dir_open NLIB_CAPI(nlib_dir_open)
#define nlib_dir_close NLIB_CAPI(nlib_dir_close)
#define nlib_dir_read NLIB_CAPI(nlib_dir_read)
#define nlib_is_dir NLIB_CAPI(nlib_is_dir)
#define nlib_exist_path NLIB_CAPI(nlib_exist_path)
#define nlib_disk_freespace NLIB_CAPI(nlib_disk_freespace)
#define nlib_basename NLIB_CAPI(nlib_basename)
#define nlib_dirname NLIB_CAPI(nlib_dirname)
#define nlib_mkostemps NLIB_CAPI(nlib_mkostemps)
#define nlib_fd_fileid NLIB_CAPI(nlib_fd_fileid)
#define nlib_readlink NLIB_CAPI(nlib_readlink)
#define nlib_spinlock_lock_ NLIB_CAPI(nlib_spinlock_lock_)
#define nlib_spinlock_unlock_ NLIB_CAPI(nlib_spinlock_unlock_)
#define nlib_vsnprintf NLIB_CAPI(nlib_vsnprintf)
#define nlib_snprintf NLIB_CAPI(nlib_snprintf)
#define nlib_vdprintf NLIB_CAPI(nlib_vdprintf)
#define nlib_dprintf NLIB_CAPI(nlib_dprintf)
#define nlib_printf NLIB_CAPI(nlib_printf)
#define nlib_vsnwprintf NLIB_CAPI(nlib_vsnwprintf)
#define nlib_snwprintf NLIB_CAPI(nlib_snwprintf)
#define nlib_vdwprintf NLIB_CAPI(nlib_vdwprintf)
#define nlib_dwprintf NLIB_CAPI(nlib_dwprintf)
#define nlib_wprintf NLIB_CAPI(nlib_wprintf)
#define nlib_vsnprintf_fallback NLIB_CAPI(nlib_vsnprintf_fallback)
#define nlib_snprintf_fallback NLIB_CAPI(nlib_snprintf_fallback)
#define nlib_vsnwprintf_fallback NLIB_CAPI(nlib_vsnwprintf_fallback)
#define nlib_snwprintf_fallback NLIB_CAPI(nlib_snwprintf_fallback)
#define nlib_memcmp NLIB_CAPI(nlib_memcmp)
#define nlib_memchr NLIB_CAPI(nlib_memchr)
#define nlib_memrchr NLIB_CAPI(nlib_memrchr)
#define nlib_memchr_not NLIB_CAPI(nlib_memchr_not)
#define nlib_memchr_range_not NLIB_CAPI(nlib_memchr_range_not)
#define nlib_memchr_lt NLIB_CAPI(nlib_memchr_lt)
#define nlib_memchr_gt NLIB_CAPI(nlib_memchr_gt)
#define nlib_memchr_mb NLIB_CAPI(nlib_memchr_mb)
#define nlib_memspn NLIB_CAPI(nlib_memspn)
#define nlib_memcspn NLIB_CAPI(nlib_memcspn)
#define nlib_memccpy NLIB_CAPI(nlib_memccpy)
#define nlib_skipws NLIB_CAPI(nlib_skipws)
#define nlib_strlen NLIB_CAPI(nlib_strlen)
#define nlib_strnlen NLIB_CAPI(nlib_strnlen)
#define nlib_strcpy NLIB_CAPI(nlib_strcpy)
#define nlib_strncpy NLIB_CAPI(nlib_strncpy)
#define nlib_strchr NLIB_CAPI(nlib_strchr)
#define nlib_strrchr NLIB_CAPI(nlib_strrchr)
#define nlib_strchr_mb NLIB_CAPI(nlib_strchr_mb)
#define nlib_wcslen NLIB_CAPI(nlib_wcslen)
#define nlib_wcsnlen NLIB_CAPI(nlib_wcsnlen)
#define nlib_wcscpy NLIB_CAPI(nlib_wcscpy)
#define nlib_wcsncpy NLIB_CAPI(nlib_wcsncpy)
#define nlib_strcat NLIB_CAPI(nlib_strcat)
#define nlib_strncat NLIB_CAPI(nlib_strncat)
#define nlib_wcscat NLIB_CAPI(nlib_wcscat)
#define nlib_wcsncat NLIB_CAPI(nlib_wcsncat)
#define nlib_strto_int8 NLIB_CAPI(nlib_strto_int8)
#define nlib_strto_int16 NLIB_CAPI(nlib_strto_int16)
#define nlib_strto_int32 NLIB_CAPI(nlib_strto_int32)
#define nlib_strto_int64 NLIB_CAPI(nlib_strto_int64)
#define nlib_strto_uint8 NLIB_CAPI(nlib_strto_uint8)
#define nlib_strto_uint16 NLIB_CAPI(nlib_strto_uint16)
#define nlib_strto_uint32 NLIB_CAPI(nlib_strto_uint32)
#define nlib_strto_uint64 NLIB_CAPI(nlib_strto_uint64)
#define nlib_strto_double NLIB_CAPI(nlib_strto_double)
#define nlib_strto_float NLIB_CAPI(nlib_strto_float)
#define nlib_strto_int32_fallback NLIB_CAPI(nlib_strto_int32_fallback)
#define nlib_strto_int64_fallback NLIB_CAPI(nlib_strto_int64_fallback)
#define nlib_strto_uint32_fallback NLIB_CAPI(nlib_strto_uint32_fallback)
#define nlib_strto_uint64_fallback NLIB_CAPI(nlib_strto_uint64_fallback)
#define nlib_strto_double_fallback NLIB_CAPI(nlib_strto_double_fallback)
#define nlib_strto_float_fallback NLIB_CAPI(nlib_strto_float_fallback)
#define nlib_wide_to_utf8 NLIB_CAPI(nlib_wide_to_utf8)
#define nlib_utf8_to_wide NLIB_CAPI(nlib_utf8_to_wide)
#define nlib_strcplen NLIB_CAPI(nlib_strcplen)
#define nlib_memcplen NLIB_CAPI(nlib_memcplen)
#define nlib_wcscplen NLIB_CAPI(nlib_wcscplen)
#define nlib_swapendian_16 NLIB_CAPI(nlib_swapendian_16)
#define nlib_swapendian_32 NLIB_CAPI(nlib_swapendian_32)
#define nlib_swapendian_64 NLIB_CAPI(nlib_swapendian_64)
#define nlib_malloc NLIB_CAPI(nlib_malloc)
#define nlib_free NLIB_CAPI(nlib_free)
#define nlib_calloc NLIB_CAPI(nlib_calloc)
#define nlib_realloc NLIB_CAPI(nlib_realloc)
#define nlib_malloc_size NLIB_CAPI(nlib_malloc_size)
#define nlib_free_size NLIB_CAPI(nlib_free_size)
#define nlib_free_size_ NLIB_CAPI(nlib_free_size_)
#define nlib_memalign NLIB_CAPI(nlib_memalign)

#define nlib_utf16cplen_ex_ NLIB_CAPI(nlib_utf16cplen_ex_)
#define nlib_utf16cpy_ NLIB_CAPI(nlib_utf16cpy_)
#define nlib_utf16len_ NLIB_CAPI(nlib_utf16len_)
#define nlib_utf16ncpy_ NLIB_CAPI(nlib_utf16ncpy_)
#define nlib_utf16nlen_ NLIB_CAPI(nlib_utf16nlen_)
#define nlib_utf16_to_utf32char NLIB_CAPI(nlib_utf16_to_utf32char)
#define nlib_utf16_to_utf8 NLIB_CAPI(nlib_utf16_to_utf8)
#define nlib_utf32char_to_utf16 NLIB_CAPI(nlib_utf32char_to_utf16)
#define nlib_utf32char_to_utf8 NLIB_CAPI(nlib_utf32char_to_utf8)
#define nlib_utf32cplen NLIB_CAPI(nlib_utf32cplen)
#define nlib_utf32cpy_ NLIB_CAPI(nlib_utf32cpy_)
#define nlib_utf32len_ NLIB_CAPI(nlib_utf32len_)
#define nlib_utf32ncpy_ NLIB_CAPI(nlib_utf32ncpy_)
#define nlib_utf32nlen_ NLIB_CAPI(nlib_utf32nlen_)
#define nlib_utf32_to_utf8 NLIB_CAPI(nlib_utf32_to_utf8)
#define nlib_utf8_to_utf16 NLIB_CAPI(nlib_utf8_to_utf16)
#define nlib_utf8_to_utf32 NLIB_CAPI(nlib_utf8_to_utf32)
#define nlib_utf8_to_utf32char NLIB_CAPI(nlib_utf8_to_utf32char)
#define nlib_memutf8_to_utf16 NLIB_CAPI(nlib_memutf8_to_utf16)
#define nlib_memutf8_to_utf32 NLIB_CAPI(nlib_memutf8_to_utf32)
#define nlib_memutf16_to_utf8 NLIB_CAPI(nlib_memutf16_to_utf8)
#define nlib_memutf32_to_utf8 NLIB_CAPI(nlib_memutf32_to_utf8)
#define nlib_memutf8_to_wide NLIB_CAPI(nlib_memutf8_to_wide)
#define nlib_memwide_to_utf8 NLIB_CAPI(nlib_memwide_to_utf8)

#define nlib_socket NLIB_CAPI(nlib_socket)
#define nlib_bind NLIB_CAPI(nlib_bind)
#define nlib_listen NLIB_CAPI(nlib_listen)
#define nlib_accept NLIB_CAPI(nlib_accept)
#define nlib_accept_for NLIB_CAPI(nlib_accept_for)
#define nlib_connect NLIB_CAPI(nlib_connect)
#define nlib_connect_for NLIB_CAPI(nlib_connect_for)
#define nlib_send NLIB_CAPI(nlib_send)
#define nlib_sendto NLIB_CAPI(nlib_sendto)
#define nlib_sendmsg NLIB_CAPI(nlib_sendmsg)
#define nlib_recv NLIB_CAPI(nlib_recv)
#define nlib_recvfrom NLIB_CAPI(nlib_recvfrom)
#define nlib_recvmsg NLIB_CAPI(nlib_recvmsg)
#define nlib_closesocket NLIB_CAPI(nlib_closesocket)
#define nlib_shutdownsocket NLIB_CAPI(nlib_shutdownsocket)
#define nlib_inet_pton NLIB_CAPI(nlib_inet_pton)
#define nlib_inet_ntop NLIB_CAPI(nlib_inet_ntop)
#define nlib_getaddrinfo NLIB_CAPI(nlib_getaddrinfo)
#define nlib_freeaddrinfo NLIB_CAPI(nlib_freeaddrinfo)
#define nlib_getsockopt NLIB_CAPI(nlib_getsockopt)
#define nlib_setsockopt NLIB_CAPI(nlib_setsockopt)
#define nlib_select NLIB_CAPI(nlib_select)
#define nlib_poll NLIB_CAPI(nlib_poll)
#define nlib_getnameinfo NLIB_CAPI(nlib_getnameinfo)
#define nlib_getsockname NLIB_CAPI(nlib_getsockname)
#define nlib_getpeername NLIB_CAPI(nlib_getpeername)
#define nlib_setnonblocking NLIB_CAPI(nlib_setnonblocking)

// not in the public header
#define nlib_utf8_nbytes_table NLIB_CAPI(nlib_utf8_nbytes_table)
#define nlib_hex_char_table NLIB_CAPI(nlib_hex_char_table)
#define nlib_memcmp_generic NLIB_CAPI(nlib_memcmp_generic)
#define nlib_memchr_generic NLIB_CAPI(nlib_memchr_generic)
#define nlib_memrchr_generic NLIB_CAPI(nlib_memrchr_generic)
#define nlib_memchr_not_generic NLIB_CAPI(nlib_memchr_not_generic)
#define nlib_memchr_range_not_generic NLIB_CAPI(nlib_memchr_range_not_generic)
#define nlib_memchr_lt_generic NLIB_CAPI(nlib_memchr_lt_generic)
#define nlib_memchr_gt_generic NLIB_CAPI(nlib_memchr_gt_generic)
#define nlib_memchr_mb_generic NLIB_CAPI(nlib_memchr_mb_generic)
#define nlib_strlen_generic NLIB_CAPI(nlib_strlen_generic)
#define nlib_skipws_generic NLIB_CAPI(nlib_skipws_generic)
#define nlib_strnlen_generic NLIB_CAPI(nlib_strnlen_generic)
#define nlib_strchr_generic NLIB_CAPI(nlib_strchr_generic)
#define nlib_strrchr_generic NLIB_CAPI(nlib_strrchr_generic)
#define nlib_swapendian_16_generic NLIB_CAPI(nlib_swapendian_16_generic)
#define nlib_swapendian_32_generic NLIB_CAPI(nlib_swapendian_32_generic)
#define nlib_swapendian_64_generic NLIB_CAPI(nlib_swapendian_64_generic)
#define nlib_utf16len_generic NLIB_CAPI(nlib_utf16len_generic)
#define nlib_utf16nlen_generic NLIB_CAPI(nlib_utf16nlen_generic)
#define nlib_utf32len_generic NLIB_CAPI(nlib_utf32len_generic)
#define nlib_utf32nlen_generic NLIB_CAPI(nlib_utf32nlen_generic)
#define nlib_utf16cplen_ex_generic NLIB_CAPI(nlib_utf16cplen_ex_generic)
#define nlib_fd_impl NLIB_CAPI(nlib_fd_impl)

#define utf16chr_mb_simd NLIB_CAPI(utf16chr_mb_simd)
#define nlib_strchr_simd NLIB_CAPI(nlib_strchr_simd)
#define nlib_strrchr_simd NLIB_CAPI(nlib_strrchr_simd)
#define nlib_memchr_simd NLIB_CAPI(nlib_memchr_simd)
#define nlib_memrchr_simd NLIB_CAPI(nlib_memrchr_simd)
#define nlib_memchr_not_simd NLIB_CAPI(nlib_memchr_not_simd)
#define nlib_memchr_range_not_simd NLIB_CAPI(nlib_memchr_range_not_simd)
#define nlib_memchr_lt_simd NLIB_CAPI(nlib_memchr_lt_simd)
#define nlib_memchr_gt_simd NLIB_CAPI(nlib_memchr_gt_simd)
#define nlib_memchr_mb_simd NLIB_CAPI(nlib_memchr_mb_simd)
#define nlib_skipws_simd NLIB_CAPI(nlib_skipws_simd)
#define nlib_strlen_simd NLIB_CAPI(nlib_strlen_simd)
#define nlib_strnlen_simd NLIB_CAPI(nlib_strnlen_simd)
#define nlib_utf16len_simd NLIB_CAPI(nlib_utf16len_simd)
#define nlib_utf16nlen_simd NLIB_CAPI(nlib_utf16nlen_simd)
#define nlib_utf16cplen_ex_simd NLIB_CAPI(nlib_utf16cplen_ex_simd)
#define nlib_memcmp_simd NLIB_CAPI(nlib_memcmp_simd)
#define nlib_swapendian_16_simd NLIB_CAPI(nlib_swapendian_16_simd)
#define nlib_swapendian_32_simd NLIB_CAPI(nlib_swapendian_32_simd)
#define nlib_swapendian_64_simd NLIB_CAPI(nlib_swapendian_64_simd)

#ifdef __NX__
#define nlib_mount_host_nx NLIB_CAPI(nlib_mount_host_nx)
#endif
#else
#ifdef NLIB_CAPI
#undef NLIB_CAPI
#endif
#define NLIB_CAPI(x) x
#endif

#define NLIB_STRINGIFY_(s) #s
#define NLIB_STRINGIFY(s) NLIB_STRINGIFY_(s)

#if defined(_MSC_VER)
# include "nn/nlib/Platform_win32.h"
#elif defined(__linux__)   || \
      defined(__FreeBSD__) || \
      defined(__CYGWIN__)  || \
     (defined(__APPLE__) && defined(__MACH__))
# ifndef NLIB_UNIX
#  define NLIB_UNIX
# endif
# include "nn/nlib/Platform_unix.h"
#elif defined(NN_PLATFORM_CTR)
# include "nn/nlib/Platform_ctr.h"
#elif defined(CAFE)
# include "nn/nlib/Platform_cafe.h"
#elif defined(__NX__)
# include "nn/nlib/Platform_nx.h"
#endif

#if defined(_MSC_VER) && defined(n_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

#ifndef __analysis_assume
# define __analysis_assume(expr)
#endif

#ifndef _Printf_format_string_
# define _Printf_format_string_
#endif

#if defined(__ARM_NEON__) || defined(__aarch64__)
# ifndef NLIB_NEON
#  define NLIB_NEON
# endif
#endif

#ifdef __SSE4_1__
# ifndef NLIB_SSE41
#  define NLIB_SSE41
# endif
#endif

#ifdef __SSE4_2__
# ifndef NLIB_SSE41
#  define NLIB_SSE41
# endif
# ifndef NLIB_SSE42
#  define NLIB_SSE42
# endif
#endif

#if defined(NLIB_SSE41) || defined(NLIB_NEON)
# define NLIB_SIMD
# ifdef NLIB_NEON
#  include <arm_neon.h>  // NOLINT
# endif
# ifdef NLIB_SSE41
#  include <smmintrin.h>  // NOLINT
# endif
# ifdef NLIB_SSE42
#  include <nmmintrin.h>  // NOLINT
# endif
#endif

#if defined(__ARM_ACLE)
#include <arm_acle.h>
#endif

// https://www.jpcert.or.jp/sc-rules/c-int01-c.html
// https://www.securecoding.cert.org/confluence/display/seccode/INT01-C.+Use+rsize_t+or+size_t+for+all+integer+values+representing+the+size+of+an+object
#ifndef RSIZE_MAX
# ifndef NLIB_64BIT
#  define RSIZE_MAX 0x7FFFFFFFL
# else
#  define RSIZE_MAX 0x7FFFFFFFFFFFFFFFLL
# endif
#endif

#ifndef NLIB_WARN
# define NLIB_WARN(exp) ("WARNING: " exp)
// #pragma message NLIB_WARN("your message")
#endif

// NOTE:
// You can use those macros on the environment which printf does not support '%z'
// for size_t.
// "%"PRIuS, sizet_val
// See: https://google-styleguide.googlecode.com/svn/trunk/cppguide.html#64-bit_Portability
// In Japanese: http://www.textdrop.net/google-styleguide-ja/cppguide.xml
#ifndef __PRIS_PREFIX
# define __PRIS_PREFIX "z"
#endif

#ifndef PRIdS
# define PRIdS __PRIS_PREFIX "d"
#endif

#ifndef PRIxS
# define PRIxS __PRIS_PREFIX "x"
#endif

#ifndef PRIuS
# define PRIuS __PRIS_PREFIX "u"
#endif

#ifndef PRIXS
# define PRIXS __PRIS_PREFIX "X"
#endif

#ifndef PRIoS
# define PRIoS __PRIS_PREFIX "o"
#endif

#ifndef NLIB_ASSUME
# define NLIB_ASSUME(cond) switch (0) case 0: default: if (cond) ; else __builtin_unreachable()  /* NOLINT */
#endif

#ifdef NLIB_NONNULL_ENABLED
# define NLIB_EINVAL_IFNULL(p)
#else
# define NLIB_EINVAL_IFNULL(p) if (!p) return EINVAL
#endif

#ifndef NLIB_C_INLINE
# define NLIB_C_INLINE __inline
#endif

//
// Error Type
//
typedef int errno_t;  // TR 24731-1

//
// stdint
//
#include <stdint.h>  // NOLINT

#ifdef NLIB_HAS_STDHEADER_INTTYPES
# include <inttypes.h>  // NOLINT
#endif

#if defined(__LP64__) && __LP64__ == 1
// if long, unsinged long are 64 bits long
#define NLIB_LP64
typedef int64_t nlib_long_compatible_t;
typedef uint64_t nlib_ulong_compatible_t;
#else
typedef int32_t nlib_long_compatible_t;
typedef uint32_t nlib_ulong_compatible_t;
#endif

#ifndef NLIB_VIS_PUBLIC_ALT
#define NLIB_VIS_PUBLIC_ALT NLIB_VIS_PUBLIC
#endif

#ifndef NLIB_EXPIMP_TEMPLATE
# define NLIB_EXPIMP_TEMPLATE(x) NLIB_STATIC_ASSERT(sizeof(char) == 1)
#endif

#ifdef __cplusplus
extern "C" {
#endif

NLIB_VIS_PUBLIC const char* nlib_error_string(errno_t e);

//
// Native error
//
NLIB_VIS_PUBLIC unsigned int nlib_get_native_last_error(void);

//
// Version
//

// returns NLIB_VERSION
NLIB_VIS_PUBLIC int nlib_getversion(void);
NLIB_VIS_PUBLIC int nlib_compiler_version(void);

// crc32 = 0 at the beginning
NLIB_VIS_PUBLIC uint32_t nlib_crc32(uint32_t crc32, const void* p, size_t n);
// crc32 = 0 at the beginning
NLIB_VIS_PUBLIC uint32_t nlib_crc32c(uint32_t crc32c, const void* p, size_t n);

//
// Atomic
//
#if defined(__INTELLISENSE__)
#define NLIB_ATOMIC_RELAXED (0)
#define NLIB_ATOMIC_ACQUIRE (1)
#define NLIB_ATOMIC_RELEASE (2)
#define NLIB_ATOMIC_ACQ_REL (3)
#define NLIB_ATOMIC_SEQ_CST (7)

int32_t nlib_atomic_load32(const int32_t* ptr, int memorder);
void nlib_atomic_store32(int32_t* ptr, int32_t val, int memorder);
// *target = value, and returns the old value of *target
int32_t nlib_atomic_exchange32(int32_t* ptr,
                               int32_t val, int memorder);
// *ptr = desired and return non-zero if successful
int nlib_atomic_compare_exchange32(int32_t* ptr,
                                   int32_t* expected,
                                   int32_t desired,
                                   int weak,
                                   int success_memorder,
                                   int failure_memorder);
// *ptr += val; return *ptr;
int32_t nlib_atomic_add_fetch32(int32_t* ptr, int32_t val, int memorder);
// *ptr -= val; return *ptr;
int32_t nlib_atomic_sub_fetch32(int32_t* ptr, int32_t val, int memorder);
// *ptr &= val; return *ptr;
int32_t nlib_atomic_and_fetch32(int32_t* ptr, int32_t val, int memorder);
// *ptr ^= val; return *ptr;
int32_t nlib_atomic_xor_fetch32(int32_t* ptr, int32_t val, int memorder);
// *ptr |= val; return *ptr;
int32_t nlib_atomic_or_fetch32(int32_t* ptr, int32_t val, int memorder);
// tmp = *ptr; *ptr += val; return tmp;
int32_t nlib_atomic_fetch_add32(int32_t* ptr, int32_t val, int memorder);
// tmp = *ptr; *ptr -= val; return tmp;
int32_t nlib_atomic_fetch_sub32(int32_t* ptr, int32_t val, int memorder);
// tmp = *ptr; *ptr &= val; return tmp;
int32_t nlib_atomic_fetch_and32(int32_t* ptr, int32_t val, int memorder);
// tmp = *ptr; *ptr ^= val; return tmp;
int32_t nlib_atomic_fetch_xor32(int32_t* ptr, int32_t val, int memorder);
// tmp = *ptr; *ptr |= val; return tmp;
int32_t nlib_atomic_fetch_or32(int32_t* ptr, int32_t val, int memorder);

int64_t nlib_atomic_load64(const int64_t* ptr, int memorder);
void nlib_atomic_store64(int64_t* ptr, int64_t val, int memorder);
// *target = value, and returns the old value of *target
int64_t nlib_atomic_exchange64(int64_t* ptr, int64_t val, int memorder);
// *ptr = desired and return non-zero if successful
int nlib_atomic_compare_exchange64(int64_t* ptr, int64_t* expected,
                                   int64_t desired, int weak,
                                   int success_memorder, int failure_memorder);
// *ptr += val; return *ptr;
int64_t nlib_atomic_add_fetch64(int64_t* ptr, int64_t val, int memorder);
// *ptr -= val; return *ptr;
int64_t nlib_atomic_sub_fetch64(int64_t* ptr, int64_t val, int memorder);
// *ptr &= val; return *ptr;
int64_t nlib_atomic_and_fetch64(int64_t* ptr, int64_t val, int memorder);
// *ptr ^= val; return *ptr;
int64_t nlib_atomic_xor_fetch64(int64_t* ptr, int64_t val, int memorder);
// *ptr |= val; return *ptr;
int64_t nlib_atomic_or_fetch64(int64_t* ptr, int64_t val, int memorder);
// tmp = *ptr; *ptr += val; return tmp;
int64_t nlib_atomic_fetch_add64(int64_t* ptr, int64_t val, int memorder);
// tmp = *ptr; *ptr -= val; return tmp;
int64_t nlib_atomic_fetch_sub64(int64_t* ptr, int64_t val, int memorder);
// tmp = *ptr; *ptr &= val; return tmp;
int64_t nlib_atomic_fetch_and64(int64_t* ptr, int64_t val, int memorder);
// tmp = *ptr; *ptr ^= val; return tmp;
int64_t nlib_atomic_fetch_xor64(int64_t* ptr, int64_t val, int memorder);
// tmp = *ptr; *ptr |= val; return tmp;
int64_t nlib_atomic_fetch_or64(int64_t* ptr, int64_t val, int memorder);

void* nlib_atomic_loadptr(void* const* ptr, int memorder);
void nlib_atomic_storeptr(void** ptr, void* val, int memorder);
void* nlib_atomic_exchangeptr(void** ptr, void* val, int memorder);
// *ptr = desired and return non-zero if successful
int nlib_atomic_compare_exchangeptr(void** ptr, void** expected, void* desired,
                                    int weak, int success_memorder, int failure_memorder);

void nlib_atomic_thread_fence(int memorder);
#endif

//
// Time, Duration
//
#ifndef NLIB_TIMESPEC_HAS_NATIVE
struct timespec {
    time_t tv_sec;
    long tv_nsec;  // NOLINT
};
#endif

// 100ns => 1, 1970/01/01 == 0
typedef int64_t nlib_time;
// 100ns => 1, 1ms => 10000
typedef int64_t nlib_duration;

// 100ns => 1, 1970/01/01 == 0
NLIB_VIS_PUBLIC errno_t nlib_epochtime(nlib_time* t) NLIB_NONNULL;
// 100ns => 1, boot time == 0, msec = *t / 10000
NLIB_VIS_PUBLIC errno_t nlib_ticktime(nlib_duration* t) NLIB_NONNULL;
// 100ns => 1, sleep 1 msec = nlib_sleep(10000)
NLIB_VIS_PUBLIC errno_t nlib_sleep(nlib_duration t);

#define NLIB_TO_TIMESPEC(tm, t) \
    (tm)->tv_sec = (time_t)((t) / (1000 * 10000)); \
    (tm)->tv_nsec = ((long)((t) % (1000 * 10000)) * 100)  // NOLINT

#define NLIB_FROM_TIMESPEC(tm, t) \
    t = (int64_t)((tm)->tv_sec) * (1000 * 10000) + ((tm)->tv_nsec / 100)

static NLIB_C_INLINE errno_t nlib_epochtime_timespec(struct timespec* tm) {
    nlib_time t;
    errno_t e = nlib_epochtime(&t);
    if (NLIB_UNLIKELY(e != 0)) return e;
    NLIB_TO_TIMESPEC(tm, t);
    return 0;
}

static NLIB_C_INLINE errno_t nlib_ticktime_timespec(struct timespec* tm) {
    nlib_duration d;
    errno_t e = nlib_ticktime(&d);
    if (NLIB_UNLIKELY(e != 0)) return e;
    NLIB_TO_TIMESPEC(tm, d);
    return 0;
}

static NLIB_C_INLINE errno_t nlib_sleep_timespec(const struct timespec* tm) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_sleep(d);
}

#if !defined(NLIB_TIMER_HAS_NATIVE)
typedef uint32_t nlib_timer;
#elif defined(_MSC_VER)
typedef HANDLE nlib_timer;
#elif defined(__linux__)
typedef int nlib_timer;
#else
# error sorry
#endif
typedef void (*nlib_timer_callback)(nlib_timer timer, void* param);
struct nlib_timerspec_ {
    nlib_duration due_time;
    nlib_duration interval;
};
typedef struct nlib_timerspec_ nlib_timerspec;
#ifdef NLIB_DOXYGEN
struct nlib_timerspec {
    nlib_duration due_time;
    nlib_duration interval;
};
#endif
NLIB_VIS_PUBLIC errno_t nlib_timer_create(nlib_timer* timer, nlib_timer_callback callback,
                                          void* param, uint32_t flags);
NLIB_VIS_PUBLIC errno_t nlib_timer_settime(nlib_timer timer, const nlib_timerspec* new_value,
                                           nlib_timerspec* old_value);
NLIB_VIS_PUBLIC errno_t nlib_timer_gettime(nlib_timer timer, nlib_timerspec* curr_value);
NLIB_VIS_PUBLIC errno_t nlib_timer_delete(nlib_timer timer,
                                          int wait_completion,
                                          nlib_timer_callback completion_callback);
#define NLIB_TIMER_SHORTTERM_TASK 0x00000001
#define NLIB_TIMER_LONGTERM_TASK 0x00000002

//
// Random
//

// Store 'size' bytes of random values on 'buf'
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_gen_random(void* buf, size_t size) NLIB_NONNULL;

//
// Virtual Memory, Physical Memory
//
NLIB_VIS_PUBLIC errno_t nlib_mempagesize(size_t* size) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_virtual_alloc(void** ptr, size_t size) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_virtual_free(void* ptr, size_t size) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_physical_alloc(void* ptr, size_t size, int prot) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_physical_free(void* ptr, size_t size) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_mlock(void* addr, size_t len) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_munlock(void* addr, size_t len) NLIB_NONNULL;

#define NLIB_PHYSICAL_ALLOC_PROT_NONE 0
#define NLIB_PHYSICAL_ALLOC_PROT_READ 1
#define NLIB_PHYSICAL_ALLOC_PROT_WRITE 2
#define NLIB_PHYSICAL_ALLOC_PROT_EXEC 4

//
// TLS
//
typedef void (*nlib_tls_destructor)(void* tls_value);
#define NLIB_TLS_INVALID (nlib_tls)(-1)

#ifdef NLIB_PTHREAD_nlib_tls_alloc
static
#else
NLIB_VIS_PUBLIC
#endif
// code snippets:
//   # map tls_key on thread local storage
//     nlib_tls tls_key;
//     e = nlib_tls_alloc(&tls_key, NULL);  # no dtor invoked if destr is NULL
//     if (e != 0) { error ... }
//   # unmap tls_key
//     nlib_tls_free(tls_key);
//   # access from a thread
//     void* thread_local_value;
//     nlib_tls_getvalue(tls_key, &thread_local_value);
//     # use and update thread_local_value
//     nlib_tls_setvalue(tls_key, thread_local_value);
NLIB_CHECK_RESULT
errno_t nlib_tls_alloc(nlib_tls* tls, nlib_tls_destructor destr) NLIB_NONNULL_1;
#ifdef NLIB_PTHREAD_nlib_tls_alloc
static NLIB_C_INLINE errno_t nlib_tls_alloc(nlib_tls* tls, nlib_tls_destructor destr) {
    return pthread_key_create(tls, destr);
}
#endif
#ifdef NLIB_PTHREAD_nlib_tls_free
static NLIB_C_INLINE errno_t nlib_tls_free(nlib_tls tls) {
    return pthread_key_delete(tls);
}
#else
NLIB_VIS_PUBLIC errno_t nlib_tls_free(nlib_tls tls);
#endif
#ifdef NLIB_PTHREAD_nlib_tls_setvalue
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_tls_setvalue(nlib_tls tls, const void* value);
#ifdef NLIB_PTHREAD_nlib_tls_setvalue
static NLIB_C_INLINE errno_t nlib_tls_setvalue(nlib_tls tls, const void* value) {
    return pthread_setspecific(tls, value);
}
#endif

#ifdef NLIB_PTHREAD_nlib_tls_getvalue
static
#else
NLIB_VIS_PUBLIC
#endif

errno_t nlib_tls_getvalue(nlib_tls tls, void** value) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_tls_getvalue
static NLIB_C_INLINE errno_t nlib_tls_getvalue(nlib_tls tls, void** value) {
    *value = pthread_getspecific(tls);
    return 0;
}
#endif

//
// Mutex
//

#ifdef NLIB_PTHREAD_nlib_mutex_init
static
#else
NLIB_VIS_PUBLIC
#endif

// you can use NLIB_MUTEX_INITIALIZER static initializer
errno_t nlib_mutex_init(nlib_mutex* mutex) NLIB_NONNULL NLIB_EXCLUDES(*mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_init
static NLIB_C_INLINE errno_t nlib_mutex_init(nlib_mutex* mutex) NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_mutex_init(mutex, NULL);
}
#endif

// you can use NLIB_RECURSIVE_MUTEX_INITIALIZER static initializer
NLIB_VIS_PUBLIC errno_t nlib_mutex_recursive_init(nlib_mutex* mutex)
    NLIB_NONNULL NLIB_EXCLUDES(*mutex);
// you can use NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER static initializer
NLIB_VIS_PUBLIC errno_t nlib_mutex_recursive_timed_init(nlib_mutex* mutex)
    NLIB_NONNULL NLIB_EXCLUDES(*mutex);

#ifdef NLIB_PTHREAD_nlib_mutex_lock
static
#else
NLIB_VIS_PUBLIC
#endif
// code snippets:
//   nlib_mutex m;
//   if (nlib_mutex_init(&m) != 0) { ... } # always returns 0 on almost all platforms?
//   nlib_mutex_lock(&m);
//      ....
//   nlib_mutex_unlock(&m);
//   nlib_mutex_destroy(&m);
errno_t nlib_mutex_lock(nlib_mutex* mutex) NLIB_NONNULL NLIB_ACQUIRE(*mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_lock
static NLIB_C_INLINE errno_t nlib_mutex_lock(nlib_mutex* mutex) NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_mutex_lock(mutex);
}
#endif

// returns EBUSY if a lock cannot be acquired
#ifdef NLIB_PTHREAD_nlib_mutex_trylock
static
#else
NLIB_VIS_PUBLIC
#endif
NLIB_CHECK_RESULT
errno_t nlib_mutex_trylock(nlib_mutex* mutex) NLIB_NONNULL NLIB_TRY_ACQUIRE(0, *mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_trylock
static NLIB_C_INLINE errno_t nlib_mutex_trylock(nlib_mutex* mutex) NLIB_TRY_ACQUIRE(0, *mutex) {
    return pthread_mutex_trylock(mutex);
}
#endif
// returns ETIMEDOUT if timeout
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mutex_trylock_for(nlib_mutex* mutex,
                               nlib_duration delta) NLIB_NONNULL NLIB_TRY_ACQUIRE(0, *mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_unlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_mutex_unlock(nlib_mutex* mutex) NLIB_NONNULL NLIB_RELEASE(*mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_unlock
static NLIB_C_INLINE errno_t nlib_mutex_unlock(nlib_mutex* mutex) NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_mutex_unlock(mutex);
}
#endif

#ifdef NLIB_PTHREAD_nlib_mutex_destroy
static
#else
NLIB_VIS_PUBLIC
#endif
// don't forget to write this, some platforms require this called.
errno_t nlib_mutex_destroy(nlib_mutex* mutex) NLIB_NONNULL NLIB_EXCLUDES(*mutex);
#ifdef NLIB_PTHREAD_nlib_mutex_destroy
static NLIB_C_INLINE errno_t nlib_mutex_destroy(nlib_mutex* mutex) NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_mutex_destroy(mutex);
}
#endif

static NLIB_C_INLINE errno_t nlib_mutex_trylock_for_timespec(nlib_mutex* mutex,
                                                             const struct timespec* tm)
    NLIB_TRY_ACQUIRE(0, *mutex) {
    nlib_duration delta;
    NLIB_FROM_TIMESPEC(tm, delta);
    return nlib_mutex_trylock_for(mutex, delta);
}

//
// Semaphore
//

NLIB_VIS_PUBLIC errno_t nlib_semaphore_init(nlib_semaphore* sem, int initial_count) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_semaphore_wait(nlib_semaphore* sem) NLIB_NONNULL;
// returns EAGAIN if semaphore cannot be acquired
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_semaphore_trywait(nlib_semaphore* sem) NLIB_NONNULL;
// returns ETIMEDOUT if timeout
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_semaphore_trywait_for(
    nlib_semaphore* sem, nlib_duration duration) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_semaphore_post(nlib_semaphore* __restrict sem,
                                            int* __restrict previous_count) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC errno_t nlib_semaphore_post_ex(nlib_semaphore* __restrict sem, int release_count,
                                               int* __restrict previous_count) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC errno_t nlib_semaphore_destroy(nlib_semaphore* sem) NLIB_NONNULL;

static NLIB_C_INLINE errno_t nlib_semaphore_trywait_for_timespec(
    nlib_semaphore* sem, const struct timespec* tm) {
    nlib_duration duration;
    NLIB_FROM_TIMESPEC(tm, duration);
    return nlib_semaphore_trywait_for(sem, duration);
}

//
// Condition Variable
//

#ifdef NLIB_PTHREAD_nlib_cond_init
static
#else
NLIB_VIS_PUBLIC
#endif

// you can use NLIB_COND_INITIALIZER for static initializer
errno_t nlib_cond_init(nlib_cond* cond) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_cond_init
static NLIB_C_INLINE errno_t nlib_cond_init(nlib_cond* cond) {
    return pthread_cond_init(cond, NULL);
}
#endif

#ifdef NLIB_PTHREAD_nlib_cond_signal
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_cond_signal(nlib_cond* cond) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_cond_signal
static NLIB_C_INLINE errno_t nlib_cond_signal(nlib_cond* cond) {
    return pthread_cond_signal(cond);
}
#endif

#ifdef NLIB_PTHREAD_nlib_cond_broadcast
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_cond_broadcast(nlib_cond* cond) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_cond_broadcast
static NLIB_C_INLINE errno_t nlib_cond_broadcast(nlib_cond* cond) {
    return pthread_cond_broadcast(cond);
}
#endif

#ifdef NLIB_PTHREAD_nlib_cond_wait
static
#else
NLIB_VIS_PUBLIC
#endif
// code snippets:
//   Initialization:
//     bool flag = false;
//     nlib_mutex m;
//     nlib_cond cond;
//     nlib_mutex_init(&m);
//     nlib_cond_init(&cond);
//   Thread1:
//     nlib_mutex_lock(&m);
//     while (!flag)
//       e = nlib_cond_wait(&cond, &m);  # m to be unlocked in nlib_cond_wait
//       # note that nlib_cond_wait may return without signal notified
//       if (e != 0) { error .... }
//     # do job and reset flag
//     flag = false;
//     nlib_mutex_unlock(&m);
//   Thread2:
//     nlib_mutex_lock(&m);
//     flag = true;
//     nlib_cond_broadcast(&cond);
//     nlib_mutex_unlock(&m);
errno_t nlib_cond_wait(nlib_cond* __restrict cond, nlib_mutex* __restrict mutex)
    NLIB_NONNULL NLIB_REQUIRES(*mutex);
#ifdef NLIB_PTHREAD_nlib_cond_wait
static NLIB_C_INLINE
errno_t nlib_cond_wait(nlib_cond* __restrict cond, nlib_mutex* __restrict mutex) {
    return pthread_cond_wait(cond, mutex);
}
#endif

// returns ETIMEDOUT if timeout, and see baloon for nlib_cond_wait()
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_cond_wait_for(nlib_cond* __restrict cond,
                           nlib_mutex* __restrict mutex,
                           nlib_duration duration) NLIB_NONNULL NLIB_REQUIRES(*mutex);
// returns ETIMEDOUT if timeout, and see baloon for nlib_cond_wait()
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_cond_wait_until(nlib_cond* __restrict cond,
                             nlib_mutex* __restrict mutex,
                             nlib_time abstime) NLIB_NONNULL NLIB_REQUIRES(*mutex);

#ifdef NLIB_PTHREAD_nlib_cond_destroy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_cond_destroy(nlib_cond* cond) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_cond_destroy
static NLIB_C_INLINE errno_t nlib_cond_destroy(nlib_cond* cond) {
    return pthread_cond_destroy(cond);
}
#endif

static NLIB_C_INLINE
errno_t nlib_cond_wait_for_timespec(nlib_cond* cond, nlib_mutex* mutex, const struct timespec* tm)
    NLIB_REQUIRES(*mutex) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_cond_wait_for(cond, mutex, d);
}

static NLIB_C_INLINE
errno_t nlib_cond_wait_until_timespec(nlib_cond* cond, nlib_mutex* mutex,
                                      const struct timespec* tm) NLIB_REQUIRES(*mutex) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_cond_wait_until(cond, mutex, d);
}

//
// Read/Write lock
//
#ifndef NLIB_RWLOCK_HAS_NATIVE
struct nlib_rwlock_ {
    int32_t _0[3];
    nlib_mutex _1[2];
    nlib_cond _2;
};
NLIB_CAPABILITY("mutex")
typedef struct nlib_rwlock_ nlib_rwlock;

#define NLIB_RWLOCK_INITIALIZER {                                                       \
    { 0, 0, 0 },                                                                        \
    { NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER, NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER }, \
      NLIB_COND_INITIALIZER }
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_init
static
#else
NLIB_VIS_PUBLIC
#endif

errno_t nlib_rwlock_init(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_EXCLUDES(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_init
static NLIB_C_INLINE errno_t nlib_rwlock_init(nlib_rwlock* rwlock) {
    return pthread_rwlock_init(rwlock, NULL);
}
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_destroy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_destroy(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_EXCLUDES(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_destroy
static NLIB_C_INLINE errno_t nlib_rwlock_destroy(nlib_rwlock* rwlock) {
    return pthread_rwlock_destroy(rwlock);
}
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_rdlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_rdlock(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_ACQUIRE_SHARED(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_rdlock
static NLIB_C_INLINE errno_t nlib_rwlock_rdlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_rdlock(rwlock);
}
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_tryrdlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_tryrdlock(nlib_rwlock* rwlock)
    NLIB_NONNULL NLIB_TRY_ACQUIRE_SHARED(0, *rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_tryrdlock
static NLIB_C_INLINE errno_t nlib_rwlock_tryrdlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_tryrdlock(rwlock);
}
#endif

NLIB_VIS_PUBLIC
errno_t nlib_rwlock_tryrdlock_for(nlib_rwlock* rwlock, nlib_duration duration)
    NLIB_NONNULL NLIB_TRY_ACQUIRE_SHARED(0, *rwlock);
NLIB_VIS_PUBLIC
errno_t nlib_rwlock_tryrdlock_until(nlib_rwlock* rwlock, nlib_time abstime)
    NLIB_NONNULL NLIB_TRY_ACQUIRE_SHARED(0, *rwlock);

#ifdef NLIB_PTHREAD_nlib_rwlock_rdunlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_rdunlock(nlib_rwlock* rwlock)
    NLIB_NONNULL NLIB_RELEASE_SHARED(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_rdunlock
static NLIB_C_INLINE errno_t nlib_rwlock_rdunlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_unlock(rwlock);
}
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_wrlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_wrlock(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_ACQUIRE(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_wrlock
static NLIB_C_INLINE errno_t nlib_rwlock_wrlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_wrlock(rwlock);
}
#endif

#ifdef NLIB_PTHREAD_nlib_rwlock_trywrlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_trywrlock(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_TRY_ACQUIRE(0, *rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_trywrlock
static NLIB_C_INLINE errno_t nlib_rwlock_trywrlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_trywrlock(rwlock);
}
#endif

NLIB_VIS_PUBLIC
errno_t nlib_rwlock_trywrlock_for(nlib_rwlock* rwlock, nlib_duration duration)
    NLIB_NONNULL NLIB_TRY_ACQUIRE(0, *rwlock);
NLIB_VIS_PUBLIC
errno_t nlib_rwlock_trywrlock_until(nlib_rwlock* rwlock, nlib_time abstime)
    NLIB_NONNULL NLIB_TRY_ACQUIRE(0, *rwlock);

#ifdef NLIB_PTHREAD_nlib_rwlock_wrunlock
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_rwlock_wrunlock(nlib_rwlock* rwlock) NLIB_NONNULL NLIB_RELEASE(*rwlock);
#ifdef NLIB_PTHREAD_nlib_rwlock_wrunlock
static NLIB_C_INLINE errno_t nlib_rwlock_wrunlock(nlib_rwlock* rwlock)
    NLIB_NO_THREAD_SAFETY_ANALYSIS {
    return pthread_rwlock_unlock(rwlock);
}
#endif

static NLIB_C_INLINE
errno_t nlib_rwlock_tryrdlock_for_timespec(nlib_rwlock* rwlock, const struct timespec* tm)
    NLIB_TRY_ACQUIRE_SHARED(0, *rwlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_rwlock_tryrdlock_for(rwlock, d);
}

static NLIB_C_INLINE
errno_t nlib_rwlock_tryrdlock_until_timespec(nlib_rwlock* rwlock, const struct timespec* tm)
    NLIB_TRY_ACQUIRE_SHARED(0, *rwlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_rwlock_tryrdlock_until(rwlock, d);
}

static NLIB_C_INLINE
errno_t nlib_rwlock_trywrlock_for_timespec(nlib_rwlock* rwlock, const struct timespec* tm)
    NLIB_TRY_ACQUIRE(0, *rwlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_rwlock_trywrlock_for(rwlock, d);
}

static NLIB_C_INLINE
errno_t nlib_rwlock_trywrlock_until_timespec(nlib_rwlock* rwlock, const struct timespec* tm)
    NLIB_TRY_ACQUIRE(0, *rwlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_rwlock_trywrlock_until(rwlock, d);
}


#if defined(_MSC_VER) && defined(NLIB_RWLOCK_HAS_NATIVE)
typedef struct nlib_condrwlock_ {
    CONDITION_VARIABLE cond;
} nlib_condrwlock;
#define NLIB_CONDRWLOCK_INITIALIZER { CONDITION_VARIABLE_INIT }
#else
typedef struct nlib_condrwlock_ {
    nlib_cond cond;
    nlib_mutex mutex;
} nlib_condrwlock;
#define NLIB_CONDRWLOCK_INITIALIZER { NLIB_COND_INITIALIZER, NLIB_MUTEX_INITIALIZER }
#endif

NLIB_VIS_PUBLIC errno_t nlib_condrwlock_init(nlib_condrwlock* cond) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_destroy(nlib_condrwlock* cond) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_signal(nlib_condrwlock* cond) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_broadcast(nlib_condrwlock* cond) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_wait(nlib_condrwlock* __restrict cond,
                                             nlib_rwlock* __restrict rwlock,
                                             int rdlock) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_wait_for(nlib_condrwlock* __restrict cond,
                                                 nlib_rwlock* __restrict rwlock,
                                                 nlib_duration duration,
                                                 int rdlock) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_condrwlock_wait_until(nlib_condrwlock* __restrict cond,
                                                   nlib_rwlock* __restrict rwlock,
                                                   nlib_time abstime,
                                                   int rdlock) NLIB_NONNULL;

static NLIB_C_INLINE errno_t nlib_condrwlock_wait_for_timespec(nlib_condrwlock* cond,
                                                               nlib_rwlock* rwlock,
                                                               const struct timespec* tm,
                                                               int rdlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_condrwlock_wait_for(cond, rwlock, d, rdlock);
}

static NLIB_C_INLINE errno_t nlib_condrwlock_wait_until_timespec(nlib_condrwlock* cond,
                                                                 nlib_rwlock* rwlock,
                                                                 const struct timespec* tm,
                                                                 int rdlock) {
    nlib_duration d;
    NLIB_FROM_TIMESPEC(tm, d);
    return nlib_condrwlock_wait_until(cond, rwlock, d, rdlock);
}

//
// Barrier
//
#ifndef NLIB_BARRIER_HAS_NATIVE
struct nlib_barrier_ {
    nlib_mutex _0;
    nlib_cond _1;
    unsigned int _2[3];
};
typedef struct nlib_barrier_ nlib_barrier;
#endif

#ifdef NLIB_PTHREAD_nlib_barrier_init
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_barrier_init(nlib_barrier* barrier, unsigned int count) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_barrier_init
static NLIB_C_INLINE errno_t nlib_barrier_init(nlib_barrier* barrier, unsigned int count) {
    return pthread_barrier_init(barrier, NULL, count);
}
#endif

#ifdef NLIB_PTHREAD_nlib_barrier_destroy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_barrier_destroy(nlib_barrier* barrier) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_barrier_destroy
static NLIB_C_INLINE errno_t nlib_barrier_destroy(nlib_barrier* barrier) {
    return pthread_barrier_destroy(barrier);
}
#endif

NLIB_VIS_PUBLIC errno_t nlib_barrier_wait(nlib_barrier* barrier) NLIB_NONNULL;

//
// Once
//
struct nlib_onceflag_ {
    int status;
};
typedef struct nlib_onceflag_ nlib_onceflag;
#define NLIB_ONCE_INIT { 0 }
typedef void (*nlib_oncefunc)(void);

// code snippets
//   void OnceFunc() { .... }
//   nlib_onceflag flag = NLIB_ONCE_INIT;  // should be static initialized
//   nlib_once(&flag, OnceFunc);  // OnceFunc executes only once
NLIB_VIS_PUBLIC errno_t nlib_once(nlib_onceflag* flag,
                                  nlib_oncefunc func) NLIB_NONNULL;
// returns EBUSY immedidately if an other thread is executing 'func'
NLIB_VIS_PUBLIC NLIB_DEPRECATED errno_t nlib_tryonce(nlib_onceflag* flag,
                                     nlib_oncefunc func) NLIB_NONNULL;

//
// Message Queue
//
#ifdef NLIB_DOXYGEN
typedef int32_t nlib_mq;
#else
typedef struct nlib_mq_ {
    int32_t raw_handle;  // 0 for invalid handle
} nlib_mq;
#endif
typedef void* nlib_mq_msg;

#define NLIB_MQ_BLOCK 0
#define NLIB_MQ_NONBLOCK 1
#define NLIB_MQ_LOCKFREE 2

typedef void (*nlib_mq_msg_destructor)(nlib_mq_msg);
#ifdef NLIB_DOXYGEN
struct nlib_mq_attr {
    int32_t flag;
    int32_t max_msg;
    int32_t cur_msg;
    nlib_mq_msg_destructor destructor;
};
#else
typedef struct nlib_mq_attr_ {
    int32_t flag;  // NLIB_MQ_BLOCK / NLIB_MQ_NONBLOCK
    int32_t max_msg;  // 0 for 128
    int32_t cur_msg;  // blocking mode only
    nlib_mq_msg_destructor destructor;
} nlib_mq_attr;
#endif

NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_open(nlib_mq* mq, const nlib_mq_attr* attr) NLIB_NONNULL;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_getattr(nlib_mq mq, nlib_mq_attr* attr) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_mq_close(nlib_mq mq);
NLIB_VIS_PUBLIC errno_t nlib_mq_readonly(nlib_mq mq);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_mq_send(nlib_mq mq, nlib_mq_msg msg, int prio);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_send_until(nlib_mq mq, nlib_mq_msg msg, int prio, nlib_time abstime);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_receive(nlib_mq mq, nlib_mq_msg* msg, int* prio) NLIB_NONNULL_2;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_receive_until(nlib_mq mq, nlib_mq_msg* msg, int* prio, nlib_time abstime)
    NLIB_NONNULL_2;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_mq_drop(nlib_mq mq, nlib_mq_msg* msg, int* prio) NLIB_NONNULL_2;

//
// Thread
//
NLIB_VIS_PUBLIC errno_t nlib_yield(void);
static NLIB_ALWAYS_INLINE void nlib_pause(void) {
#if defined(__x86_64__) || defined(__i386__) || defined(_M_IX86) || defined(_M_AMD64)
    _mm_pause();
#elif defined(__ARM_ACLE)
    __yield();
#else
    (void)nlib_yield();
#endif
}

#define NLIB_THREAD_INVALID (nlib_thread)(0)  // NOLINT

#ifndef NLIB_SPINLOCK_HAS_NATIVE
typedef int32_t nlib_spinlock;
#endif

#ifndef NLIB_THREAD_ATTR_HAS_NATIVE
struct nlib_thread_attr_ {
    nlib_spinlock spin;
    int detach_state;
    int explicit_sched;
    int priority;
    uint32_t affinity;
    void* stack_addr;
    size_t stack_size;
};
typedef struct nlib_thread_attr_ nlib_thread_attr;
#else
struct nlib_thread_attr_ {
    pthread_attr_t attr;
    uint32_t affinity;
};
typedef struct nlib_thread_attr_ nlib_thread_attr;
#endif
typedef void (*nlib_thread_func)(void* arg);
// -1 for invalid
typedef int nlib_thread_id;

// code snippets:
//   nlib_thread th;
//   if ((e = nlib_thread_create(&th, NULL, myfunc, myarg)) != 0) { error ... }
//   nlib_thread_join(th); # or nlib_thread_detach(th);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_thread_create(nlib_thread* __restrict thread, const nlib_thread_attr* __restrict attr,
                           nlib_thread_func func, void* __restrict arg)
    NLIB_NONNULL_1 NLIB_NONNULL_3;
#ifdef NLIB_PTHREAD_nlib_thread_join
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_thread_join(nlib_thread thread);
#ifdef NLIB_PTHREAD_nlib_thread_join
static NLIB_C_INLINE errno_t nlib_thread_join(nlib_thread thread) {
    return pthread_join(thread, NULL);
}
#endif

#ifdef NLIB_PTHREAD_nlib_thread_detach
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_thread_detach(nlib_thread thread);
#ifdef NLIB_PTHREAD_nlib_thread_detach
static NLIB_C_INLINE errno_t nlib_thread_detach(nlib_thread thread) {
    return pthread_detach(thread);
}
#endif

#ifdef NLIB_PTHREAD_nlib_thread_self
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_thread_self(nlib_thread* thread) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_thread_self
static NLIB_C_INLINE errno_t nlib_thread_self(nlib_thread* thread) {
    *thread = pthread_self();
    return 0;
}
#endif

NLIB_VIS_PUBLIC errno_t nlib_thread_getconcurrency(unsigned int* num_cpu) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_getid(nlib_thread_id* id) NLIB_NONNULL;

#ifdef NLIB_PTHREAD_nlib_thread_equal
static
#else
NLIB_VIS_PUBLIC
#endif
int nlib_thread_equal(nlib_thread th1, nlib_thread th2);
#ifdef NLIB_PTHREAD_nlib_thread_equal
static NLIB_C_INLINE int nlib_thread_equal(nlib_thread th1, nlib_thread th2) {
    return pthread_equal(th1, th2);
}
#endif

NLIB_VIS_PUBLIC errno_t nlib_thread_getcpu(int* result) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_setaffinity(nlib_thread thread, uint32_t affinity);
NLIB_VIS_PUBLIC errno_t nlib_thread_setname(nlib_thread thread, const char* name) NLIB_NONNULL;

#ifdef NLIB_PTHREAD_nlib_thread_getname
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_thread_getname(nlib_thread thread, char* name, size_t len) NLIB_NONNULL;
#ifdef NLIB_PTHREAD_nlib_thread_getname
static NLIB_C_INLINE errno_t nlib_thread_getname(nlib_thread thread, char* name, size_t len) {
    return pthread_getname_np(thread, name, len);
}
#endif

// NOTE:
// win32 does not have GetThreadAffinityMask()
// errno_t nlib_thread_get_affinify(nlib_thread thread, uint32_t* affinity);

NLIB_VIS_PUBLIC errno_t nlib_thread_attr_init(nlib_thread_attr* attr) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_setint(nlib_thread_attr* attr, int key, int value) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_getint(const nlib_thread_attr* __restrict attr, int key,
                                int* __restrict value) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_setptr(nlib_thread_attr* __restrict attr, int key,
                                void* __restrict value) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_getptr(const nlib_thread_attr* __restrict attr, int key,
                                void** __restrict value) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_setstack(nlib_thread_attr* __restrict attr, void* __restrict stack_addr,
                                  size_t stack_size) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_thread_attr_getstack(const nlib_thread_attr* __restrict attr,
                                  void** __restrict stack_addr, size_t* __restrict stack_size)
    NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_attr_destroy(nlib_thread_attr* attr) NLIB_NONNULL;

#define NLIB_THREAD_ATTR_KEY_DETACHSTATE    (1)
#define NLIB_THREAD_ATTR_KEY_STACKSIZE      (2)
#define NLIB_THREAD_ATTR_KEY_PRIORITY       (4)
#define NLIB_THREAD_ATTR_KEY_AFFINITY       (5)
#define NLIB_THREAD_ATTR_KEY_EXPLICIT_SCHED (6)

NLIB_VIS_PUBLIC errno_t nlib_thread_getpriority(nlib_thread thread, int* priority) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_setpriority(nlib_thread thread, int priority);
NLIB_VIS_PUBLIC errno_t nlib_thread_priority_min(int* priority) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_priority_max(int* priority) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_thread_priority_default(int* priority) NLIB_NONNULL;

#ifndef NN_PLATFORM_CTR
// See also nlib_thread_exit_cpp();
NLIB_VIS_PUBLIC void nlib_thread_exit(void) NLIB_NORETURN;
#endif

#ifdef NLIB_DOXYGEN
void nlib_thread_cleanup_push(void (*fn)(void*), void* arg);
void nlib_thread_cleanup_pop(int exec);
#elif defined(pthread_cleanup_push)
# define nlib_thread_cleanup_push(fn, arg) pthread_cleanup_push(fn, arg)
# define nlib_thread_cleanup_pop(exec) pthread_cleanup_pop(exec)
#elif !defined(NN_PLATFORM_CTR)
struct nlib_thread_cleanup_handler_ {
    void (*func)(void*);
    void* arg;
    struct nlib_thread_cleanup_handler_* next;
};
#define nlib_thread_cleanup_push(fn, arg) switch (0) case 0: default: { \
    struct nlib_thread_cleanup_handler_ _thread_cleanup_handler = { fn, arg, NULL }; \
    nlib_thread_cleanup_push_(&_thread_cleanup_handler)
#define nlib_thread_cleanup_pop(exec) nlib_thread_cleanup_pop_(exec); }

NLIB_VIS_PUBLIC void nlib_thread_cleanup_push_(struct nlib_thread_cleanup_handler_* handler);
NLIB_VIS_PUBLIC void nlib_thread_cleanup_pop_(int exec);
#endif

//
// Console/Debug
//

// note that buf is not null terminated
NLIB_VIS_PUBLIC
errno_t nlib_write_stdout(size_t* __restrict result, const void* __restrict buf, size_t count)
    NLIB_NONNULL;
// note that buf is not null terminated
NLIB_VIS_PUBLIC
errno_t nlib_write_stderr(size_t* __restrict result, const void* __restrict buf, size_t count)
    NLIB_NONNULL;
NLIB_VIS_PUBLIC void nlib_debug_break(void);
NLIB_VIS_PUBLIC
errno_t nlib_debug_backtrace(size_t* __restrict result, void** __restrict buffer, size_t count)
    NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_debug_backtrace_gettext(char* __restrict str, size_t strbufsize,
                                     void* const* __restrict buf, size_t count) NLIB_NONNULL;
NLIB_VIS_PUBLIC
errno_t nlib_getenv(size_t* __restrict result, char* __restrict buf, size_t bufsize,
                    const char* __restrict varname) NLIB_NONNULL_1 NLIB_NONNULL_4;

typedef enum nlib_log_priority {
    NLIB_LOG_UNKNOWN = 0,
    NLIB_LOG_DEAFULT,
    NLIB_LOG_VERBOSE,
    NLIB_LOG_DEBUG,
    NLIB_LOG_INFO,
    NLIB_LOG_WARN,
    NLIB_LOG_ERROR,
    NLIB_LOG_FATAL,
    NLIB_LOG_SILENT,
    NLIB_LOG_LEVEL_EQUAL_OR_ABOVE = 0x10,
    NLIB_LOG_LEVEL_EQUAL_OR_BELOW = 0x20,
    NLIB_LOG_LEVEL_ALL = 0x30
} nlib_log_priority;

typedef enum nlib_log_key {
    NLIB_LOG_ATTR_UNKNOWN = 0,
    NLIB_LOG_ATTR_STDOUT,
    NLIB_LOG_ATTR_STDERR,
    NLIB_LOG_ATTR_MSVC_TRACE,
    NLIB_LOG_ATTR_SYSLOG,
    NLIB_LOG_ATTR_NLIB_FD,
    NLIB_LOG_ATTR_MAX
} nlib_log_key;

#ifndef NLIB_ATTRIBUTE_PRINTF
# define NLIB_ATTRIBUTE_PRINTF(x, y) __attribute__((format(printf, x, y)))
#endif

NLIB_VIS_PUBLIC int nlib_log_print(int prio, _Printf_format_string_ const char* __restrict tag,
                                   const char* __restrict fmt, ...)
                                   NLIB_ATTRIBUTE_PRINTF(3, 4) NLIB_NONNULL;
NLIB_VIS_PUBLIC int nlib_log_vprint(int prio, _Printf_format_string_ const char* __restrict tag,
                                    const char* __restrict fmt, va_list ap) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_log_attr_setint(int prio, int key, int value);

//
// File Access
//
#ifndef NLIB_FD_O_RDONLY
# ifndef O_RDONLY
#  error
# endif
# define NLIB_FD_O_RDONLY O_RDONLY
#endif

#ifndef NLIB_FD_O_WRONLY
# ifndef O_WRONLY
#  error
# endif
# define NLIB_FD_O_WRONLY O_WRONLY
#endif

#ifndef NLIB_FD_O_RDWR
# ifndef O_RDWR
#  error
# endif
# define NLIB_FD_O_RDWR   O_RDWR
#endif

#ifndef NLIB_FD_O_APPEND
# ifndef O_APPEND
#  error
# endif
# define NLIB_FD_O_APPEND O_APPEND
#endif

#ifndef NLIB_FD_O_CREAT
# ifndef O_CREAT
#  error
# endif
# define NLIB_FD_O_CREAT  O_CREAT
#endif

#ifndef NLIB_FD_O_TRUNC
# ifndef O_TRUNC
#  error
# endif
# define NLIB_FD_O_TRUNC  O_TRUNC
#endif

#ifndef NLIB_FD_O_EXCL
# ifndef O_EXCL
#  error
# endif
# define NLIB_FD_O_EXCL   O_EXCL
#endif

#ifndef NLIB_SEEK_SET
# ifndef SEEK_SET
#  error
# endif
# define NLIB_SEEK_SET SEEK_SET
#endif

#ifndef NLIB_SEEK_CUR
# ifndef SEEK_CUR
#  error
# endif
# define NLIB_SEEK_CUR SEEK_CUR
#endif

// NOTE:
// SEEK_END not supported(because of FIO19-C)

typedef int64_t nlib_offset;
typedef int nlib_fd;
#define NLIB_FD_INVALID (-1)

#ifdef NLIB_DOXYGEN
errno_t nlib_fd_open(nlib_fd* fd, const char* native_path, unsigned int flags);
errno_t nlib_fd_open(nlib_fd* fd, const char* native_path, unsigned int flags, int mode);
#else
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_open(nlib_fd* fd, const char* native_path, unsigned int flags, ...) NLIB_NONNULL_1;
#endif
NLIB_CHECK_RESULT static NLIB_C_INLINE
errno_t nlib_fd_creat(nlib_fd* fd, const char* native_path, int mode) {
    return nlib_fd_open(fd, native_path,
        NLIB_FD_O_CREAT | NLIB_FD_O_WRONLY | NLIB_FD_O_EXCL, mode);
}
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_fd_close(nlib_fd fd);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_read(size_t* __restrict result, nlib_fd fd, void* __restrict buf, size_t count)
    NLIB_NONNULL_1;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_write(size_t* __restrict result, nlib_fd fd, const void* __restrict buf,
                      size_t count) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_seek(nlib_offset* result, nlib_fd fd, nlib_offset offset, int whence)
     NLIB_NONNULL_1;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_pread(size_t* __restrict result, nlib_fd fd, void* __restrict buf,
                      size_t count, nlib_offset offset) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_pwrite(size_t* __restrict result, nlib_fd fd, const void* __restrict buf,
                       size_t count, nlib_offset offset) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_fd_truncate(nlib_fd fd, nlib_offset length);

NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_getsize(nlib_offset* size, nlib_fd fd) NLIB_NONNULL;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_fd_flush(nlib_fd fd);
NLIB_VIS_PUBLIC errno_t nlib_fd_sync(nlib_fd fd);
NLIB_VIS_PUBLIC errno_t nlib_fd_native_handle(void** native_handle, nlib_fd fd) NLIB_NONNULL;
// errno_t nlib_fd_fcntl_getflag(unsigned int* flags, nlib_fd fd);
// errno_t nlib_fd_fcntl_setflag(nlib_fd fd, unsigned int flags);

// Scatter/Gather buffer
#if !defined(NLIB_IOVEC_HAS_NATIVE)
struct nlib_fd_iovec_ {
    void* iov_base;
    size_t iov_len;
};
typedef struct nlib_fd_iovec_ nlib_fd_iovec;
#else
typedef struct iovec nlib_fd_iovec;
#endif
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_readv(size_t* __restrict result, nlib_fd fd, const nlib_fd_iovec* __restrict iov,
                      int iovcnt) NLIB_NONNULL;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_writev(size_t* __restrict result, nlib_fd fd, const nlib_fd_iovec* __restrict iov,
                       int iovcnt) NLIB_NONNULL;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_preadv(size_t* __restrict result, nlib_fd fd, const nlib_fd_iovec* __restrict iov,
                       int iovcnt, nlib_offset offset) NLIB_NONNULL;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_fd_pwritev(size_t* __restrict result, nlib_fd fd, const nlib_fd_iovec* __restrict iov,
                        int iovcnt, nlib_offset offset) NLIB_NONNULL;
// errno_t nlib_fd_stat(stat* stat, nlib_fd fd);

NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_unlink(const char* native_path);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_mkdir(const char* native_path,
                                                     unsigned int flags);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_rmdir(const char* native_path);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT errno_t nlib_remove(const char* native_path);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_rename(const char* __restrict old_path, const char* __restrict new_path);

struct nlib_dir_ {
    void* ptr;
};
typedef struct nlib_dir_ nlib_dir;
typedef struct nlib_dirent_ {
    uint32_t flags;  // 0: file, 1: directory
    char name[768];
} nlib_dirent;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_dir_open(nlib_dir* __restrict dir, const char* __restrict native_path) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC errno_t nlib_dir_close(nlib_dir dir);
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_dir_read(nlib_dirent* ent, nlib_dir dir) NLIB_NONNULL;

// CTR does not have getcwd.....
// errno_t nlib_getcwd(char* buf, size_t bufsize);
// errno_t nlib_chdir(const char* path);

NLIB_VIS_PUBLIC
errno_t nlib_is_dir(int* __restrict result, const char* __restrict native_path) NLIB_NONNULL_1;
// *result != 0 if exists
NLIB_VIS_PUBLIC
errno_t nlib_exist_path(int* __restrict result, const char* __restrict native_path) NLIB_NONNULL_1;
NLIB_VIS_PUBLIC
errno_t nlib_disk_freespace(const char* __restrict native_path,
                            uint64_t* __restrict free_bytes_available,
                            uint64_t* __restrict total_bytes,
                            uint64_t* __restrict total_free_bytes);

NLIB_VIS_PUBLIC const char* nlib_basename(const char* path) NLIB_NONNULL;
NLIB_VIS_PUBLIC const char* nlib_dirname(size_t* len, const char* path) NLIB_NONNULL;
NLIB_VIS_PUBLIC errno_t nlib_mkostemps(nlib_fd* fd, char* templ, int suffixlen, int flags);

struct nlib_fileid_ {
#if defined(_MSC_VER)
    FILE_ID_INFO _;
#elif defined(NLIB_UNIX)
    dev_t _0;
    ino_t _1;
#else
    uint64_t _;
#endif
};
typedef struct nlib_fileid_ nlib_fileid;
NLIB_VIS_PUBLIC errno_t nlib_fd_fileid(nlib_fileid* result, nlib_fd fd);
NLIB_VIS_PUBLIC errno_t nlib_readlink(size_t* len, const char* native_path, char* buf,
                                      size_t bufsize);

//
// Socket(Win32/Linux/Cygwin only)
//
#if defined(_MSC_VER) || defined(NLIB_UNIX)
#define NLIB_SOCKET_ENABLED

// sizeof(SOCKET) is 8 on Win64, but it's safe because the upper 32bit of SOCKET is always 0.
// https://stackoverflow.com/questions/1953639/is-it-safe-to-cast-socket-to-int-under-win64
typedef int nlib_sock;
#define NLIB_SOCKET_INVALID (nlib_sock)(-1)  // NOLINT

#if defined(_MSC_VER)
#define NLIB_SOCK_NONBLOCK 0x8000
#elif defined(SOCK_NONBLOCK)
#define NLIB_SOCK_NONBLOCK SOCK_NONBLOCK
#else
#define NLIB_SOCK_NONBLOCK O_NONBLOCK
#endif

#if !defined(TCP_FASTOPEN) && defined(_MSC_VER)
// (older) Windows SDK may not define TCP_FASTOPEN even if available
# define TCP_FASTOPEN 15
#endif

NLIB_VIS_PUBLIC errno_t nlib_socket(nlib_sock* sockfd, int af, int type, int protocol);

typedef struct sockaddr nlib_sockaddr;
typedef struct sockaddr_in nlib_sockaddr_in;
typedef struct sockaddr_in6 nlib_sockaddr_in6;
#ifdef _MSC_VER
struct nlib_msghdr_ {
    void* msg_name;
    uint32_t msg_namelen;
    nlib_fd_iovec* msg_iov;
    size_t msg_iovlen;
    void* msg_control;
    size_t msg_controllen;
    int msg_flags;
};
typedef struct nlib_msghdr_ nlib_msghdr;
#define NLIB_CMSG_FIRSTHDR(msgh) \
    ((msgh)->msg_controllen >= sizeof(WSACMSGHDR) ? \
    (LPWSACMSGHDR)(msgh)->msg_control : (LPWSACMSGHDR)NULL)
#define NLIB_CMSG_NXTHDR(msg, cmsg)                       \
    ( ((cmsg) == NULL) ? NLIB_CMSG_FIRSTHDR(msg)          \
        : ( ( ((PUCHAR)(cmsg) +                           \
                    WSA_CMSGHDR_ALIGN((cmsg)->cmsg_len) + \
                    sizeof(WSACMSGHDR) ) >                \
                (PUCHAR)((msg)->msg_control) +            \
                    (msg)->msg_controllen)                \
            ? (LPWSACMSGHDR)NULL                          \
            : (LPWSACMSGHDR)((PUCHAR)(cmsg) +             \
                WSA_CMSGHDR_ALIGN((cmsg)->cmsg_len)) ) )
#define NLIB_CMSG_SPACE(length) WSA_CMSG_SPACE(length)
#define NLIB_CMSG_LEN(length) WSA_CMSG_LEN(length)
#define NLIB_CMSG_DATA(cmsg) WSA_CMSG_DATA(cmsg)
#else
typedef struct msghdr nlib_msghdr;
#define NLIB_CMSG_FIRSTHDR(msgh) CMSG_FIRSTHDR(msgh)
#define NLIB_CMSG_NXTHDR(msgh, cmsg) CMSG_NXTHDR(msgh, cmsg)
// #define NLIB_CMSG_ALIGN(length) CMSG_ALIGN(length)
#define NLIB_CMSG_SPACE(length) CMSG_SPACE(length)
#define NLIB_CMSG_LEN(length) CMSG_LEN(length)
#define NLIB_CMSG_DATA(cmsg) CMSG_DATA(cmsg)
#endif
typedef struct cmsghdr nlib_cmsghdr;
NLIB_VIS_PUBLIC errno_t nlib_bind(nlib_sock sockfd, const nlib_sockaddr* addr, uint32_t namelen);
NLIB_VIS_PUBLIC errno_t nlib_listen(nlib_sock sockfd, int backlog);
NLIB_VIS_PUBLIC
errno_t nlib_accept(nlib_sock* __restrict s, nlib_sock sockfd, nlib_sockaddr* __restrict addr,
                    uint32_t* __restrict addrlen, int flags);
NLIB_VIS_PUBLIC
errno_t nlib_accept_for(nlib_sock* __restrict s, nlib_sock sockfd, nlib_sockaddr* __restrict addr,
                        uint32_t* __restrict addrlen, int flags, nlib_duration timeout);
NLIB_VIS_PUBLIC
errno_t nlib_connect(nlib_sock sockfd, const nlib_sockaddr* addr, uint32_t addrlen);
NLIB_VIS_PUBLIC
errno_t nlib_connect_for(nlib_sock sockfd, const nlib_sockaddr* addr, uint32_t addrlen,
                         int flags, nlib_duration timeout);

NLIB_VIS_PUBLIC
errno_t nlib_sendto(size_t* __restrict size, nlib_sock sockfd, const void* __restrict buf,
                    size_t len, int flags, const nlib_sockaddr* __restrict dest_addr,
                    uint32_t addrlen);
NLIB_VIS_PUBLIC
errno_t nlib_sendmsg(size_t* __restrict size, nlib_sock sockfd, const nlib_msghdr* msg, int flags);
static NLIB_C_INLINE
errno_t nlib_send(size_t* __restrict size, nlib_sock sockfd, const void* __restrict buf,
                  size_t len, int flags) {
    return nlib_sendto(size, sockfd, buf, len, flags, NULL, 0);
}

NLIB_VIS_PUBLIC
errno_t nlib_recvfrom(size_t* __restrict size, nlib_sock sockfd,
                      void* __restrict buf, size_t len, int flags,
                      nlib_sockaddr* __restrict dest_addr,
                      uint32_t* __restrict addrlen);
NLIB_VIS_PUBLIC
errno_t nlib_recvmsg(size_t* __restrict size, nlib_sock sockfd, nlib_msghdr* msg, int flags);
static NLIB_C_INLINE
errno_t nlib_recv(size_t* __restrict size, nlib_sock sockfd, void* __restrict buf, size_t len,
                  int flags) {
    return nlib_recvfrom(size, sockfd, buf, len, flags, NULL, NULL);
}

NLIB_VIS_PUBLIC errno_t nlib_closesocket(nlib_sock sockfd);

#ifdef _MSC_VER
# define SHUT_RD SD_RECEIVE
# define SHUT_WR SD_SEND
# define SHUT_RDWR SD_BOTH
#endif

NLIB_VIS_PUBLIC errno_t nlib_shutdownsocket(nlib_sock sockfd, int how);

#ifndef NLIB_LITTLE_ENDIAN
static NLIB_C_INLINE uint32_t nlib_htonl(uint32_t hostlong) { return hostlong; }
static NLIB_C_INLINE uint16_t nlib_htons(uint16_t hostshort) { return hostshort; }
static NLIB_C_INLINE uint32_t nlib_ntohl(uint32_t netlong) { return netlong; }
static NLIB_C_INLINE uint16_t nlib_ntohs(uint16_t netshort) { return netshort; }
#else
static NLIB_C_INLINE uint32_t nlib_htonl(uint32_t hostlong) {
#ifdef _MSC_VER
    return _byteswap_ulong(hostlong);
#else
    return __builtin_bswap32(hostlong);
#endif
}
static NLIB_C_INLINE uint16_t nlib_htons(uint16_t hostshort) {
#ifdef _MSC_VER
    return _byteswap_ushort(hostshort);
#else
    return ((hostshort & 0xFF) << 8) | ((hostshort >> 8) & 0xFF);
#endif
}
static NLIB_C_INLINE uint32_t nlib_ntohl(uint32_t netlong) {
#ifdef _MSC_VER
    return _byteswap_ulong(netlong);
#else
    return __builtin_bswap32(netlong);
#endif
}
static NLIB_C_INLINE uint16_t nlib_ntohs(uint16_t netshort) {
#ifdef _MSC_VER
    return _byteswap_ushort(netshort);
#else
    return ((netshort & 0xFF) << 8) | ((netshort >> 8) & 0xFF);
#endif
}
#endif

typedef struct in_addr nlib_in_addr;
typedef struct in6_addr nlib_in6_addr;
NLIB_VIS_PUBLIC errno_t nlib_inet_pton(int af, const char* __restrict src, void* __restrict dst);
NLIB_VIS_PUBLIC errno_t nlib_inet_ntop(int af, const void* __restrict src, char* __restrict dst,
                                       uint32_t dst_size);

/*
# define EAI_ADDRFAMILY 1
# define EAI_AGAIN 2
# define EAI_BADFLAGS 3
# define EAI_FAIL 4
# define EAI_FAMILY 5
# define EAI_MEMORY 6
# define EAI_NODATA 7
# define EAI_NONAME 8
# define EAI_SERVICE 9
# define EAI_SOCKTYPE 10
# define EAI_SYSTEM 11
# define EAI_BADHINTS 12
# define EAI_PROTOCOL 13
# define EAI_OVERFLOW 14
*/

// for EAI_ADDRFAMILY, EAI_AGAIN, ... etc.
typedef int eai_error_t;
typedef struct addrinfo nlib_addrinfo;
NLIB_VIS_PUBLIC
eai_error_t nlib_getaddrinfo(const char* __restrict node, const char* __restrict service,
                             const nlib_addrinfo* __restrict hints,
                             nlib_addrinfo** __restrict res);
NLIB_VIS_PUBLIC
eai_error_t nlib_getnameinfo(const nlib_sockaddr* __restrict sa, uint32_t salen,
                             char* __restrict host, uint32_t hostlen,
                             char* __restrict serv, uint32_t servlen, int flags);
NLIB_VIS_PUBLIC void nlib_freeaddrinfo(nlib_addrinfo* res);

NLIB_VIS_PUBLIC
errno_t nlib_getsockopt(nlib_sock sockfd, int level, int optname, void* __restrict optval,
                        uint32_t* __restrict optlen);
NLIB_VIS_PUBLIC
errno_t nlib_setsockopt(nlib_sock sockfd, int level, int optname, const void* optval,
                        uint32_t optlen);

typedef fd_set nlib_fd_set;
NLIB_VIS_PUBLIC
errno_t nlib_select(size_t* __restrict n, int nfds, nlib_fd_set* __restrict readfds,
                    nlib_fd_set* __restrict writefds,
                    nlib_fd_set* __restrict exceptfds, nlib_duration timeout);

#define NLIB_FD_CLR FD_CLR
#define NLIB_FD_ISSET FD_ISSET
#ifdef _MSC_VER
#define NLIB_FD_SET(fd, set) FD_SET((SOCKET)fd, set)
#else
#define NLIB_FD_SET FD_SET
#endif
#define NLIB_FD_ZERO FD_ZERO

#if defined(_MSC_VER)
struct nlib_pollfd_ {
    nlib_sock fd;
    int16_t events;
    int16_t revents;
};
typedef struct nlib_pollfd_ nlib_pollfd;
#else
typedef struct pollfd nlib_pollfd;
#endif
NLIB_VIS_PUBLIC
errno_t nlib_poll(size_t* __restrict n, nlib_pollfd* __restrict fds, uint32_t nfds,
                  nlib_duration timeout);

#ifdef _MSC_VER
#define NLIB_MSG_DONTWAIT (0x4000)
#else
#define NLIB_MSG_DONTWAIT MSG_DONTWAIT
#endif

#ifdef MSG_FASTOPEN
#define NLIB_MSG_FASTOPEN MSG_FASTOPEN
#else
#define NLIB_MSG_FASTOPEN 0x20000000
#endif

// NLIB_VIS_PUBLIC errno_t nlib_sockatmark(int* is_marked, nlib_sock sockfd);
NLIB_VIS_PUBLIC
errno_t nlib_getsockname(nlib_sock sockfd, nlib_sockaddr* __restrict addr,
                         uint32_t* __restrict addrlen);
NLIB_VIS_PUBLIC
errno_t nlib_getpeername(nlib_sock sockfd, nlib_sockaddr* __restrict addr,
                         uint32_t* __restrict addrlen);
NLIB_VIS_PUBLIC errno_t nlib_setnonblocking(nlib_sock sockfd, int nonblock);

#endif

//
// errno workaround
//

// PLEASE DO NOT DEFINE POSIX.1-2008 errno macros in your code
// http://pubs.opengroup.org/onlinepubs/9699919799/
#ifdef _MSC_VER
#  ifdef EDQUOT
#   warning  Do not define EDQUOT. nlib may not work correctly.
#  endif
#endif

/*
#ifdef NN_PLATFORM_CTR
// socket_User.autogen.h defines POSIX.1-2008 errno macros as enum values.
# ifdef NN_SOCKET_SOCKET_USER_AUTOGEN_H_
#  warning DO NOT INCLUDE socket_User.autogen.h in CTR socket library, Compile may fail.
#  define NLIB_SKIP_ERRNO_DEFINE
#  ifdef __errno_h
#   warning DO NOT INCLUDE errno.h when you include socket_User.autogen.h, you may not be able to handle errors of CTR socket library correctly.  // NOLINT
#  endif
# endif
#endif
*/

#ifndef NLIB_SKIP_ERRNO_DEFINE
#include <errno.h>  // NOLINT, for POSIX error values
#define NLIB_E_BASE 2000

// if boost/cerrno.hpp has definitions on errno values,
// nlib defines the same values which boost/cerrno.hpp defines.

// [E2BIG]
//     Argument list too long.
#ifndef E2BIG
# define E2BIG 9946
#endif

// [EACCES]
//    Permission denied.
#ifndef EACCES
# define EACCES 9973
#endif

// [EADDRINUSE]
//    Address in use.
#ifndef EADDRINUSE
# define EADDRINUSE 9902
#endif

// [EADDRNOTAVAIL]
//    Address not available.
#ifndef EADDRNOTAVAIL
# define EADDRNOTAVAIL 9903
#endif

// [EAFNOSUPPORT]
//    Address family not supported.
#ifndef EAFNOSUPPORT
# define EAFNOSUPPORT 9901
#endif

// [EAGAIN]
//    Resource unavailable, try again (may be the same value as [EWOULDBLOCK]).
#ifndef EAGAIN
# define EAGAIN 9976
#endif

// [EALREADY]
//    Connection already in progress.
#ifndef EALREADY
# define EALREADY 9907
#endif

// [EBADF]
//    Bad file descriptor.
#ifndef EBADF
# define EBADF 9949
#endif

// [EBADMSG]
//    Bad message.
#ifndef EBADMSG
# define EBADMSG 9905
#endif

// [EBUSY]
//    Device or resource busy.
#ifndef EBUSY
# define EBUSY 9952
#endif

// [ECANCELED]
// Operation canceled.
#ifndef ECANCELED
# define ECANCELED 9927
#endif

// [ECHILD]
//    No child processes.
#ifndef ECHILD
# define ECHILD 9963
#endif

// [ECONNABORTED]
//    Connection aborted.
#ifndef ECONNABORTED
# define ECONNABORTED 9906
#endif

// [ECONNREFUSED]
//    Connection refused.
#ifndef ECONNREFUSED
# define ECONNREFUSED 9908
#endif

// [ECONNRESET]
//    Connection reset.
#ifndef ECONNRESET
# define ECONNRESET 9909
#endif

// [EDEADLK]
//    Resource deadlock would occur.
#ifndef EDEADLK
# define EDEADLK 9975
#endif

// [EDESTADDRREQ]
//    Destination address required.
#ifndef EDESTADDRREQ
# define EDESTADDRREQ 9910
#endif

// [EDOM]
//    Mathematics argument out of domain of function.
#ifndef EDOM
# define EDOM 9947
#endif

// [EDQUOT]
//    Reserved.
#ifndef EDQUOT
# define EDQUOT (NLIB_E_BASE + 19)
#endif

// [EEXIST]
//    File exists.
#ifndef EEXIST
# define EEXIST 9955
#endif

// [EFAULT]
//    Bad address.
#ifndef EFAULT
# define EFAULT 9948
#endif

// [EFBIG]
//    File too large.
#ifndef EFBIG
# define EFBIG 9956
#endif

// [EHOSTUNREACH]
//    Host is unreachable.
#ifndef EHOSTUNREACH
# define EHOSTUNREACH 9911
#endif

// [EIDRM]
//    Identifier removed.
#ifndef EIDRM
# define EIDRM 9912
#endif

// [EILSEQ]
//    Illegal byte sequence.
#ifndef EILSEQ
# define EILSEQ 9945
#endif

// [EINPROGRESS]
//    Operation in progress.
#ifndef EINPROGRESS
# define EINPROGRESS 9928
#endif

// [EINTR]
//    Interrupted function.
#ifndef EINTR
# define EINTR 9959
#endif

// [EINVAL]
//    Invalid argument.
#ifndef EINVAL
# define EINVAL 9943
#endif

// [EIO]
//    I/O error.
#ifndef EIO
# define EIO 9961
#endif

// [EISCONN]
//    Socket is connected.
#ifndef EISCONN
# define EISCONN 9904
#endif

// [EISDIR]
//    Is a directory.
#ifndef EISDIR
# define EISDIR 9962
#endif

// [ELOOP]
//    Too many levels of symbolic links.
#ifndef ELOOP
# define ELOOP 9939
#endif

// [EMFILE]
//    File descriptor value too large.
#ifndef EMFILE
# define EMFILE 9978
#endif

// [EMLINK]
//    Too many links.
#ifndef EMLINK
# define EMLINK 9979
#endif

// [EMSGSIZE]
//    Message too large.
#ifndef EMSGSIZE
# define EMSGSIZE 9913
#endif

// [EMULTIHOP]
//    Reserved.
#ifndef EMULTIHOP
# define EMULTIHOP (NLIB_E_BASE + 36)
#endif

// [ENAMETOOLONG]
//    Filename too long.
#ifndef ENAMETOOLONG
# define ENAMETOOLONG 9957
#endif

// [ENETDOWN]
//    Network is down.
#ifndef ENETDOWN
# define ENETDOWN 9914
#endif

// [ENETRESET]
//    Connection aborted by network.
#ifndef ENETRESET
# define ENETRESET 9915
#endif

// [ENETUNREACH]
//    Network unreachable.
#ifndef ENETUNREACH
# define ENETUNREACH 9916
#endif

// [ENFILE]
//    Too many files open in system.
#ifndef ENFILE
# define ENFILE 9977
#endif

// [ENOBUFS]
//    No buffer space available.
#ifndef ENOBUFS
# define ENOBUFS 9917
#endif

// [ENODATA]
//    [OB XSR] [Option Start] No message is available on the STREAM head read queue. [Option End]
#ifndef ENODATA
# define ENODATA 9919
#endif

// [ENODEV]
//    No such device.
#ifndef ENODEV
# define ENODEV 9967
#endif

// [ENOENT]
//    No such file or directory.
#ifndef ENOENT
# define ENOENT 9968
#endif

// [ENOEXEC]
//    Executable file format error.
#ifndef ENOEXEC
# define ENOEXEC 9954
#endif

// [ENOLCK]
//    No locks available.
#ifndef ENOLCK
# define ENOLCK 9964
#endif

// [ENOLINK]
//    Reserved.
#ifndef ENOLINK
# define ENOLINK 9918
#endif

// [ENOMEM]
//    Not enough space.
#ifndef ENOMEM
# define ENOMEM 9971
#endif

// [ENOMSG]
//    No message of the desired type.
#ifndef ENOMSG
# define ENOMSG 9920
#endif

// [ENOPROTOOPT]
//    Protocol not available.
#ifndef ENOPROTOOPT
# define ENOPROTOOPT 9921
#endif

// [ENOSPC]
//    No space left on device.
#ifndef ENOSPC
# define ENOSPC 9965
#endif

// [ENOSR]
//    [OB XSR] [Option Start] No STREAM resources. [Option End]
#ifndef ENOSR
# define ENOSR 9922
#endif

// [ENOSTR]
//    [OB XSR] [Option Start] Not a STREAM. [Option End]
#ifndef ENOSTR
# define ENOSTR 9924
#endif

// [ENOSYS]
//    Function not supported.
#ifndef ENOSYS
# define ENOSYS 9942
#endif

// [ENOTCONN]
//    The socket is not connected.
#ifndef ENOTCONN
# define ENOTCONN 9925
#endif

// [ENOTDIR]
//    Not a directory or a symbolic link to a directory.
#ifndef ENOTDIR
# define ENOTDIR 9970
#endif

// [ENOTEMPTY]
//    Directory not empty.
#ifndef ENOTEMPTY
# define ENOTEMPTY 9953
#endif

// [ENOTRECOVERABLE]
//    State not recoverable.
#ifndef ENOTRECOVERABLE
# define ENOTRECOVERABLE 9934
#endif

// [ENOTSOCK]
//    Not a socket.
#ifndef ENOTSOCK
# define ENOTSOCK 9923
#endif

// [ENOTSUP]
//    Not supported (may be the same value as [EOPNOTSUPP]).
#ifndef ENOTSUP
# define ENOTSUP 9926
#endif

// [ENOTTY]
//    Inappropriate I/O control operation.
#ifndef ENOTTY
# define ENOTTY 9958
#endif

// [ENXIO]
//    No such device or address.
#ifndef ENXIO
# define ENXIO 9966
#endif

// [EOPNOTSUPP]
//    Operation not supported on socket (may be the same value as [ENOTSUP]).
#ifndef EOPNOTSUPP
# define EOPNOTSUPP 9929
#endif

// [EOVERFLOW]
//    Value too large to be stored in data type.
#ifndef EOVERFLOW
# define EOVERFLOW 9940
#endif

// [EOWNERDEAD]
//    Previous owner died.
#ifndef EOWNERDEAD
# define EOWNERDEAD  9931
#endif

// [EPERM]
//    Operation not permitted.
#ifndef EPERM
# define EPERM 9972
#endif

// [EPIPE]
//    Broken pipe.
#ifndef EPIPE
# define EPIPE 9950
#endif

// [EPROTO]
//    Protocol error.
#ifndef EPROTO
# define EPROTO 9932
#endif

// [EPROTONOSUPPORT]
//    Protocol not supported.
#ifndef EPROTONOSUPPORT
# define EPROTONOSUPPORT 9933
#endif

// [EPROTOTYPE]
//    Protocol wrong type for socket.
#ifndef EPROTOTYPE
# define EPROTOTYPE 9941
#endif

// [ERANGE]
//    Result too large.
#ifndef ERANGE
# define ERANGE 9944
#endif

// [EROFS]
//    Read-only file system.
#ifndef EROFS
# define EROFS 9974
#endif

// [ESPIPE]
//    Invalid seek.
#ifndef ESPIPE
# define ESPIPE 9960
#endif

// [ESRCH]
//    No such process.
#ifndef ESRCH
# define ESRCH 9969
#endif

// [ESTALE]
//    Reserved.
#ifndef ESTALE
# define ESTALE (NLIB_E_BASE + 75)
#endif

// [ETIME]
//    [OB XSR] [Option Start] Stream ioctl() timeout. [Option End]
#ifndef ETIME
# define ETIME 9935
#endif

// [ETIMEDOUT]
//    Connection timed out.
#ifndef ETIMEDOUT
# define ETIMEDOUT 9938
#endif

// [ETXTBSY]
//    Text file busy.
#ifndef ETXTBSY
# define ETXTBSY 9936
#endif

// [EWOULDBLOCK]
//    Operation would block (may be the same value as [EAGAIN]).
#ifndef EWOULDBLOCK
# define EWOULDBLOCK 9930
#endif

// [EXDEV]
//    Cross-device link.
#ifndef EXDEV
# define EXDEV 9951
#endif
#endif  // NLIB_SKIP_ERRNO_DEFINE

//
// nlib_spinlock
//
#ifndef NLIB_SPINLOCK_HAS_NATIVE
#define NLIB_SPINLOCK_INITIALIZER (0)
NLIB_VIS_PUBLIC void nlib_spinlock_lock_(nlib_spinlock* lock) NLIB_NONNULL;
#ifndef NN_PLATFORM_CTR
NLIB_VIS_PUBLIC void nlib_spinlock_unlock_(nlib_spinlock* lock) NLIB_NONNULL;
#endif

static void nlib_spinlock_init(nlib_spinlock* lock) NLIB_NONNULL;
static void nlib_spinlock_lock(nlib_spinlock* lock) NLIB_NONNULL;
static errno_t nlib_spinlock_trylock(nlib_spinlock* lock) NLIB_NONNULL;
static void nlib_spinlock_unlock(nlib_spinlock* lock) NLIB_NONNULL;

static NLIB_ALWAYS_INLINE void nlib_spinlock_init(nlib_spinlock* lock) {
    *lock = 0;
}
static NLIB_ALWAYS_INLINE void nlib_spinlock_lock(nlib_spinlock* lock) {
#if defined(__arm__)
#if __has_builtin(__builtin_arm_ldrex)
    int R5 = __builtin_arm_ldrex(lock);
#else
    int R5 = __ldrex(lock);
#endif
    if (R5 == 0) {
#if __has_builtin(__builtin_arm_strex)
        if (__builtin_arm_strex(1, lock) == 0) {
#else
        if (__strex(1, lock) == 0) {
#endif
#if !defined(NN_PLATFORM_CTR)
            __dmb(0xf);
#else
            nlib_ctr_barrier();
#endif
        }
        return;
    }
    nlib_spinlock_lock_(lock);
#else
    int32_t expected = 0;
    if (!nlib_atomic_compare_exchange32(lock, &expected, 1, 1,
                                        NLIB_ATOMIC_ACQUIRE, NLIB_ATOMIC_RELAXED)) {
        nlib_spinlock_lock_(lock);
    }
#endif
}
static NLIB_ALWAYS_INLINE errno_t nlib_spinlock_trylock(nlib_spinlock* lock) {
#if defined(__arm__)
#if __has_builtin(__builtin_arm_ldrex)
    int R5 = __builtin_arm_ldrex(lock);
#else
    int R5 = __ldrex(lock);
#endif
    if (R5 == 0) {
#if __has_builtin(__builtin_arm_strex)
        if (__builtin_arm_strex(1, lock) == 0) {
#else
        if (__strex(1, lock) == 0) {
#endif
#if !defined(NN_PLATFORM_CTR)
            __dmb(0xf);
#else
            nlib_ctr_barrier();
#endif
            return 0;
        }
    }
    return EBUSY;
#else
    int32_t expected = 0;
    if (nlib_atomic_compare_exchange32(lock, &expected, 1, 0,
        NLIB_ATOMIC_ACQUIRE, NLIB_ATOMIC_RELAXED))
        return 0;
    else
        return EBUSY;
#endif
}
static NLIB_ALWAYS_INLINE void nlib_spinlock_unlock(nlib_spinlock* lock) {
#if defined(NN_PLATFORM_CTR)
    nlib_atomic_store32(lock, 0, NLIB_ATOMIC_RELEASE);
#else
    int32_t expected = 1;
    if (!nlib_atomic_compare_exchange32(lock, &expected, 0, 0,
        NLIB_ATOMIC_RELEASE, NLIB_ATOMIC_RELAXED)) {
        NLIB_ASSUME(expected == 2);
        nlib_spinlock_unlock_(lock);
    }
#if defined(__arm__)
    __sev();
#endif
#endif
}
#endif

//
// Utilities(Safer style functions)
//
NLIB_VIS_PUBLIC
errno_t nlib_vsnprintf(size_t* __restrict count, char* __restrict buf, size_t size,
                       _Printf_format_string_ const char* __restrict fmt, va_list args)
    NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_snprintf(size_t* __restrict count, char* __restrict buf, size_t size,
                      _Printf_format_string_ const char* __restrict fmt, ...)
    NLIB_ATTRIBUTE_PRINTF(4, 5) NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_vdprintf(nlib_fd fd, size_t* __restrict count,
                      _Printf_format_string_ const char* __restrict fmt, va_list args)
    NLIB_NONNULL_3;
NLIB_VIS_PUBLIC
errno_t nlib_dprintf(nlib_fd fd, size_t* __restrict count,
                     _Printf_format_string_ const char* __restrict fmt, ...)
    NLIB_ATTRIBUTE_PRINTF(3, 4) NLIB_NONNULL_3;
// fmt and string must be in UTF-8 even if you use Visual Studio
NLIB_VIS_PUBLIC
int nlib_printf(_Printf_format_string_ const char* fmt, ...)
    NLIB_ATTRIBUTE_PRINTF(1, 2) NLIB_NONNULL_1;

NLIB_VIS_PUBLIC
errno_t nlib_vsnwprintf(size_t* __restrict count, wchar_t* __restrict buf, size_t size,
                        _Printf_format_string_ const wchar_t* __restrict fmt, va_list args)
    NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_snwprintf(size_t* __restrict count, wchar_t* __restrict buf, size_t size,
                       _Printf_format_string_ const wchar_t* __restrict fmt, ...)
    NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_vdwprintf(nlib_fd fd, size_t* __restrict count,
                       _Printf_format_string_ const wchar_t* __restrict fmt, va_list args)
    NLIB_NONNULL_3;
NLIB_VIS_PUBLIC
errno_t nlib_dwprintf(nlib_fd fd, size_t* __restrict count,
                      _Printf_format_string_ const wchar_t* __restrict fmt, ...) NLIB_NONNULL_3;
NLIB_VIS_PUBLIC
int nlib_wprintf(_Printf_format_string_ const wchar_t* fmt, ...) NLIB_NONNULL_1;

NLIB_VIS_PUBLIC
errno_t nlib_vsnprintf_fallback(size_t* __restrict count, char* __restrict buf, size_t size,
                                _Printf_format_string_ const char* __restrict fmt, va_list args)
    NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_snprintf_fallback(size_t* __restrict count, char* __restrict buf,
                               size_t size, _Printf_format_string_ const char* __restrict fmt, ...)
    NLIB_ATTRIBUTE_PRINTF(4, 5) NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_vsnwprintf_fallback(size_t* __restrict count, wchar_t* __restrict buf, size_t size,
                                 _Printf_format_string_ const wchar_t* __restrict fmt, va_list args)
    NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_snwprintf_fallback(size_t* __restrict count, wchar_t* __restrict buf, size_t size,
                                _Printf_format_string_ const wchar_t* __restrict fmt, ...)
    NLIB_NONNULL_4;

// http://www.open-std.org/jtc1/sc22/wg14/www/docs/n1078.pdf
// http://www.open-std.org/jtc1/sc22/wg14/www/docs/n1079.htm
// a bit different from Microsoft's strcpy_s or etc.
// returns ERANGE instead of EINVAL

static errno_t nlib_memcpy(void* __restrict s1, size_t s1max, const void* __restrict s2, size_t n)
    NLIB_NONNULL;
static errno_t nlib_memmove(void* s1, size_t s1max, const void* s2, size_t n) NLIB_NONNULL;
static errno_t nlib_memset(void* buf, int ch, size_t n) NLIB_NONNULL;
NLIB_VIS_PUBLIC
void* nlib_memccpy(void* __restrict dest, size_t dest_size, const void* __restrict src,
                   size_t src_size, int c) NLIB_NONNULL;

#ifdef NLIB_LIBC_nlib_memcmp
static
#else
NLIB_VIS_PUBLIC_ALT
#endif
int nlib_memcmp(const void* buf1, const void* buf2, size_t n) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_memcmp
static NLIB_C_INLINE int nlib_memcmp(const void* buf1, const void* buf2, size_t n) {
    return memcmp(buf1, buf2, n);
}
#endif

NLIB_VIS_PUBLIC_ALT const void* nlib_memchr(const void* s, int c, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC_ALT const void* nlib_memrchr(const void* s, int c, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC_ALT const void* nlib_memchr_not(const void* s, int c, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC_ALT
const void* nlib_memchr_range_not(const void* __restrict s, const char* __restrict range,
                                  size_t n) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC_ALT const void* nlib_memchr_lt(const void* s, int c, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC_ALT const void* nlib_memchr_gt(const void* s, int c, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
// find (c & 0x80) != 0
NLIB_VIS_PUBLIC_ALT const void* nlib_memchr_mb(const void* s, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC size_t nlib_memspn(const void* __restrict buf, size_t len,
                                   const char* __restrict set, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
NLIB_VIS_PUBLIC size_t nlib_memcspn(const void* __restrict buf, size_t len,
                                    const char* __restrict set, size_t n)
    NLIB_NONNULL NLIB_ATTRIBUTE_PURE;

// ' ', CR, LF, HT are skipped
NLIB_VIS_PUBLIC_ALT
const char* nlib_skipws(size_t* __restrict cnt_lf, const char** __restrict last_lf,
                        const char* __restrict s, size_t n) NLIB_NONNULL_2;

#ifdef NLIB_LIBC_nlib_strlen
static
#else
NLIB_VIS_PUBLIC_ALT
#endif
size_t nlib_strlen(const char* s) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_strlen
static NLIB_C_INLINE size_t nlib_strlen(const char* s) { return strlen(s); }
#endif

#ifdef NLIB_LIBC_nlib_strnlen
static
#else
NLIB_VIS_PUBLIC_ALT
#endif
size_t nlib_strnlen(const char* s, size_t maxsize) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_strnlen
static NLIB_C_INLINE size_t nlib_strnlen(const char* s, size_t maxsize) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return strnlen_s(s, maxsize);
#else
    return strnlen(s, maxsize);
#endif
}
#endif

#ifdef NLIB_LIBC_nlib_strcpy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_strcpy(char* __restrict s1, size_t s1max, const char* __restrict s2) NLIB_NONNULL;
#ifdef NLIB_LIBC_nlib_strcpy
static NLIB_C_INLINE
errno_t nlib_strcpy(char*  __restrict s1, size_t s1max, const char* __restrict s2) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return strcpy_s(s1, s1max, s2);
#else
# error
#endif
}
#endif

#ifdef NLIB_LIBC_nlib_strncpy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_strncpy(char* __restrict s1, size_t s1max, const char* __restrict s2, size_t n)
    NLIB_NONNULL;
#ifdef NLIB_LIBC_nlib_strncpy
static NLIB_C_INLINE
errno_t nlib_strncpy(char* __restrict s1, size_t s1max, const char* __restrict s2, size_t n) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return strncpy_s(s1, s1max, s2, n);
#else
# error
#endif
}
#endif

#ifdef NLIB_LIBC_nlib_strchr
static
#else
NLIB_VIS_PUBLIC_ALT
#endif
const char* nlib_strchr(const char* s, int c) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_strchr
static NLIB_C_INLINE const char* nlib_strchr(const char* s, int c) { return strchr(s, c); }
#endif

#ifdef NLIB_LIBC_nlib_strrchr
static
#else
NLIB_VIS_PUBLIC_ALT
#endif
const char* nlib_strrchr(const char* s, int c) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_strrchr
static NLIB_C_INLINE const char* nlib_strrchr(const char* s, int c) { return strrchr(s, c); }
#endif

// find (c & 0x80) != 0, used for skipping ASCII chars
static const char* nlib_strchr_mb(const char* s) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
static NLIB_C_INLINE const char* nlib_strchr_mb(const char* s) {
    size_t n = nlib_strlen(s);
    const void* p = nlib_memchr_mb(s, n);
    if (p) {
        return (const char*)p;  // NOLINT
    } else {
        return s + n;
    }
}

#ifdef NLIB_LIBC_nlib_wcslen
static
#else
NLIB_VIS_PUBLIC
#endif
size_t nlib_wcslen(const wchar_t* s) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_wcslen
static NLIB_C_INLINE size_t nlib_wcslen(const wchar_t* s) { return wcslen(s); }
#endif

#ifdef NLIB_LIBC_nlib_wcsnlen
static
#else
NLIB_VIS_PUBLIC
#endif
size_t nlib_wcsnlen(const wchar_t* s, size_t maxsize) NLIB_NONNULL NLIB_ATTRIBUTE_PURE;
#ifdef NLIB_LIBC_nlib_wcsnlen
static NLIB_C_INLINE size_t nlib_wcsnlen(const wchar_t* s, size_t maxsize) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return wcsnlen_s(s, maxsize);
#else
    return wcsnlen(s, maxsize);
#endif
}
#endif

#ifdef NLIB_LIBC_nlib_wcscpy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_wcscpy(wchar_t* __restrict s1, size_t s1max, const wchar_t* __restrict s2)
    NLIB_NONNULL;
#ifdef NLIB_LIBC_nlib_wcscpy
static NLIB_C_INLINE
errno_t nlib_wcscpy(wchar_t* __restrict s1, size_t s1max, const wchar_t* __restrict s2) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return wcscpy_s(s1, s1max, s2);
#else
# error
#endif
}
#endif

#ifdef NLIB_LIBC_nlib_wcsncpy
static
#else
NLIB_VIS_PUBLIC
#endif
errno_t nlib_wcsncpy(wchar_t* __restrict s1, size_t s1max, const wchar_t* __restrict s2, size_t n)
    NLIB_NONNULL;
#ifdef NLIB_LIBC_nlib_wcsncpy
static NLIB_C_INLINE
errno_t nlib_wcsncpy(wchar_t* __restrict s1, size_t s1max, const wchar_t* __restrict s2, size_t n) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return wcsncpy_s(s1, s1max, s2, n);
#else
# error
#endif
}
#endif

NLIB_VIS_PUBLIC
errno_t nlib_strto_int32(int32_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_int64(int64_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_uint32(uint32_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_uint64(uint64_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_double(double* result, const char* nptr, char** endptr)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_float(float* result, const char* nptr, char** endptr)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_int32_fallback(int32_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_int64_fallback(int64_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_uint32_fallback(uint32_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_uint64_fallback(uint64_t* result, const char* nptr, char** endptr, int base)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_double_fallback(double* result, const char* nptr, char** endptr)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
NLIB_VIS_PUBLIC
errno_t nlib_strto_float_fallback(float* result, const char* nptr, char** endptr)
    NLIB_NONNULL_1 NLIB_NONNULL_2;
static NLIB_C_INLINE
errno_t nlib_strto_int8(int8_t* result, const char* nptr, char** endptr, int base) {
    int32_t tmp;
    errno_t e;
    e = nlib_strto_int32(&tmp, nptr, endptr, base);
    if (e != 0 && e != ERANGE) return e;
    if (tmp > 127 || tmp < -128) {
        *result = tmp < 0 ? -128 : 127;
        return ERANGE;
    }
    *result = (int8_t)tmp;  // NOLINT
    return e;
}
static NLIB_C_INLINE
errno_t nlib_strto_int16(int16_t* result, const char* nptr, char** endptr, int base) {
    int32_t tmp;
    errno_t e;
    e = nlib_strto_int32(&tmp, nptr, endptr, base);
    if (e != 0 && e != ERANGE) return e;
    if (tmp > 32767 || tmp < -32768) {
        *result = tmp < 0 ? -32768 : 32767;
        return ERANGE;
    }
    *result = (int16_t)tmp;  // NOLINT
    return e;
}
static NLIB_C_INLINE
errno_t nlib_strto_uint8(uint8_t* result, const char* nptr, char** endptr, int base) {
    uint32_t tmp;
    errno_t e;
    e = nlib_strto_uint32(&tmp, nptr, endptr, base);
    if (e != 0 && e != ERANGE) return e;
    if (tmp > 255) {
        *result = 255;
        return ERANGE;
    }
    *result = (uint8_t)tmp;  // NOLINT
    return e;
}
static NLIB_C_INLINE
errno_t nlib_strto_uint16(uint16_t* result, const char* nptr, char** endptr, int base) {
    uint32_t tmp;
    errno_t e;
    e = nlib_strto_uint32(&tmp, nptr, endptr, base);
    if (e != 0 && e != ERANGE) return e;
    if (tmp > 65535) {
        *result = 65535;
        return ERANGE;
    }
    *result = (uint16_t)tmp;  // NOLINT
    return e;
}

NLIB_VIS_PUBLIC
errno_t nlib_wide_to_utf8(size_t* __restrict utf8count, char* __restrict utf8,
                          size_t buflen, const wchar_t* __restrict wcstr) NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_utf8_to_wide(size_t* __restrict wccount, wchar_t* __restrict wcstr,
                          size_t buflen, const char* __restrict utf8) NLIB_NONNULL_4;
NLIB_VIS_PUBLIC
errno_t nlib_memwide_to_utf8(size_t* __restrict to_count, size_t* __restrict from_count,
                             char* __restrict to, size_t to_size,
                             const wchar_t* __restrict from, size_t from_size)
    NLIB_NONNULL_1 NLIB_NONNULL_2 NLIB_NONNULL_5;
NLIB_VIS_PUBLIC
errno_t nlib_memutf8_to_wide(size_t* __restrict to_count, size_t* __restrict from_count,
                             wchar_t* __restrict to, size_t to_size,
                             const char* __restrict from, size_t from_size)
    NLIB_NONNULL_1 NLIB_NONNULL_2 NLIB_NONNULL_5;

NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_wcscplen(size_t* __restrict count, const wchar_t* __restrict str) NLIB_NONNULL_2;
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_strcplen(size_t* __restrict codepoint_count,
                      size_t* __restrict supplementary_codepoint_count,
                      size_t* __restrict len,
                      const char* __restrict str) NLIB_NONNULL_4;
#define nlib_strcplen2 nlib_strcplen
NLIB_VIS_PUBLIC NLIB_CHECK_RESULT
errno_t nlib_memcplen(size_t* __restrict codepoint_count,
                      size_t* __restrict supplementary_codepoint_count,
                      size_t* __restrict from_read,
                      const char* __restrict from,
                      size_t from_size) NLIB_NONNULL_3 NLIB_NONNULL_4;

// for (0..count) { swapendian(p[count]); }
NLIB_VIS_PUBLIC_ALT errno_t nlib_swapendian_16(uint16_t* p, size_t count) NLIB_NONNULL;
// for (0..count) { swapendian(p[count]); }
NLIB_VIS_PUBLIC_ALT errno_t nlib_swapendian_32(uint32_t* p, size_t count) NLIB_NONNULL;
// for (0..count) { swapendian(p[count]); }
NLIB_VIS_PUBLIC_ALT errno_t nlib_swapendian_64(uint64_t* p, size_t count) NLIB_NONNULL;

//
// malloc functions which nlib uses
// You can redefine them.
// See sample replace_malloc.cpp of nlibnx_heap.a library.
//

// weak function
NLIB_VIS_PUBLIC_ALT NLIB_CHECK_RESULT
void* nlib_malloc(size_t size) NLIB_ATTRIBUTE_MALLOC NLIB_ATTRIBUTE_ALLOC_SIZE1(1);

// weak function
NLIB_VIS_PUBLIC_ALT void nlib_free(void* ptr);

// weak function
NLIB_VIS_PUBLIC_ALT NLIB_CHECK_RESULT
void* nlib_calloc(size_t nmemb, size_t size)
    NLIB_ATTRIBUTE_MALLOC NLIB_ATTRIBUTE_ALLOC_SIZE2(1, 2);

// weak function
NLIB_VIS_PUBLIC_ALT NLIB_CHECK_RESULT
void* nlib_realloc(void* ptr, size_t size) NLIB_ATTRIBUTE_MALLOC NLIB_ATTRIBUTE_ALLOC_SIZE1(2);

// weak function, not defined if CAFE or CTR
NLIB_VIS_PUBLIC size_t nlib_malloc_size(const void* ptr) NLIB_NONNULL;

// weak function(calls nlib_free(ptr) by default)
NLIB_VIS_PUBLIC_ALT void nlib_free_size(void* ptr, size_t size);

// weak function, not defined if WIN32 or CTR
NLIB_VIS_PUBLIC_ALT NLIB_CHECK_RESULT
void* nlib_memalign(size_t alignment, size_t size)
    NLIB_ATTRIBUTE_MALLOC NLIB_ATTRIBUTE_ALLOC_SIZE1(2) NLIB_ATTRIBUTE_ALLOC_ALIGN(1);

#ifndef NLIB_MEMCPY
# define NLIB_MEMCPY(a, b, c) memcpy((a), (b), (c))
#endif

#ifndef NLIB_MEMMOVE
# define NLIB_MEMMOVE(a, b, c) memmove((a), (b), (c))
#endif

#ifndef NLIB_MEMSET
# define NLIB_MEMSET(a, b, c) memset((a), (b), (c))
#endif

// ctype.h without locale
static int nlib_isalnum(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isalpha(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isblank(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_iscntrl(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isdigit(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isgraph(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_islower(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isprint(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_ispunct(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isspace(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isupper(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_isxdigit(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_tolower(int ch) NLIB_ATTRIBUTE_CONST;
static int nlib_toupper(int ch) NLIB_ATTRIBUTE_CONST;

static NLIB_C_INLINE int nlib_isalnum(int ch) {
    return ('0' <= ch && ch <= '9') || ('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z');
}
static NLIB_C_INLINE int nlib_isalpha(int ch) {
    return ('A' <= ch && ch <= 'Z') || ('a' <= ch && ch <= 'z');
}
static NLIB_C_INLINE int nlib_isblank(int ch) { return ch == ' ' || ch == '\t'; }
static NLIB_C_INLINE int nlib_iscntrl(int ch) { return (ch >= 0 && ch <= 0x1F) || ch == 0x7F; }
static NLIB_C_INLINE int nlib_isdigit(int ch) { return ('0' <= ch && ch <= '9'); }
static NLIB_C_INLINE int nlib_isgraph(int ch) { return ch >= 0x21 && ch <= 0x7E; }
static NLIB_C_INLINE int nlib_islower(int ch) { return (ch >= 'a' && ch <= 'z'); }
static NLIB_C_INLINE int nlib_isprint(int ch) { return ch >= 0x20 && ch <= 0x7E; }
static NLIB_C_INLINE int nlib_ispunct(int ch) { return (ch >= 0x00 && ch <= 0x20) || ch == 0x7F; }
static NLIB_C_INLINE int nlib_isspace(int ch) {
    return ((ch) == ' ' || (ch) == '\t' || (ch) == '\n');
}
static NLIB_C_INLINE int nlib_isupper(int ch) { return (ch >= 'A' && ch <= 'Z'); }
static NLIB_C_INLINE int nlib_isxdigit(int ch) {
    return (unsigned int)(ch - '0') < 10u ||
        (unsigned int)((ch | 0x20) - 'a') < 6u;
}
static NLIB_C_INLINE int nlib_tolower(int ch) {
    return (ch >= 'A' && ch <= 'Z') ? ch + ('a' - 'A') : ch;
}
static NLIB_C_INLINE int nlib_toupper(int ch) {
    return (ch >= 'a' && ch <= 'z') ? ch - ('a' - 'A') : ch;
}
// no isascii, toascii

// memcpy_s
static NLIB_C_INLINE
errno_t nlib_memcpy(void* __restrict s1, size_t s1max, const void* __restrict s2, size_t n) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return memcpy_s(s1, s1max, s2, n);
#else
#ifndef NLIB_NONNULL_ENABLED
    if (!s1 || !s2) return ERANGE;
#endif
    if (s1max < n) {
        NLIB_MEMSET(s1, 0, s1max);
        return ERANGE;
    }
    NLIB_MEMCPY(s1, s2, n);
    return 0;
#endif
}

// memmove_s
static NLIB_C_INLINE
errno_t nlib_memmove(void* s1, size_t s1max, const void* s2, size_t n) {
#if defined(_MSC_VER) || defined(__STDC_LIB_EXT1__)
    return memmove_s(s1, s1max, s2, n);
#else
#ifndef NLIB_NONNULL_ENABLED
    if (!s1 || !s2) return ERANGE;
#endif
    if (s1max < n) return ERANGE;
    NLIB_MEMMOVE(s1, s2, n);
    return 0;
#endif
}

static NLIB_C_INLINE
errno_t nlib_memset(void* buf, int ch, size_t n) {
    NLIB_EINVAL_IFNULL(buf);
    NLIB_MEMSET(buf, ch, n);
    return 0;
}

static int nlib_popcnt16(uint16_t x) NLIB_ATTRIBUTE_CONST;
static int nlib_popcnt32(uint32_t x) NLIB_ATTRIBUTE_CONST;
static int nlib_popcnt64(uint64_t x) NLIB_ATTRIBUTE_CONST;
#if defined(NLIB_SSE42)
static NLIB_ALWAYS_INLINE int nlib_popcnt16(uint16_t x) {
    return _mm_popcnt_u32(x);
}
static NLIB_ALWAYS_INLINE int nlib_popcnt32(uint32_t x) {
    return _mm_popcnt_u32(x);
}
static NLIB_ALWAYS_INLINE int nlib_popcnt64(uint64_t x) {
#ifdef NLIB_64BIT
    return (int)_mm_popcnt_u64(x);  // NOLINT
#else
    uint32_t lo = (uint32_t)(x & 0xFFFFFFFFU);  // NOLINT
    uint32_t hi = (uint32_t)((x >> 32) & 0xFFFFFFFFU);  // NOLINT
    return _mm_popcnt_u32(lo) + _mm_popcnt_u32(hi);  // NOLINT
#endif
}
#elif defined(NLIB_NEON)
static NLIB_ALWAYS_INLINE int nlib_popcnt16(uint16_t x) {
    uint8x8_t x0 = vcnt_u8(vreinterpret_u8_u64(vcreate_u64(x)));
#ifdef __aarch64__
    return vaddv_u8(x0);
#else
    uint8x8_t x1 = vpadd_u8(x0, x0);
    return vget_lane_u8(x1, 0);
#endif
}
static NLIB_ALWAYS_INLINE int nlib_popcnt32(uint32_t x) {
    uint8x8_t x0 = vcnt_u8(vreinterpret_u8_u64(vcreate_u64(x)));
#ifdef __aarch64__
    return vaddv_u8(x0);
#else
    uint8x8_t x1 = vpadd_u8(x0, x0);
    uint8x8_t x2 = vpadd_u8(x1, x1);
    return vget_lane_u8(x2, 0);
#endif
}
static NLIB_ALWAYS_INLINE int nlib_popcnt64(uint64_t x) {
    uint8x8_t x0 = vcnt_u8(vreinterpret_u8_u64(vcreate_u64(x)));
#ifdef __aarch64__
    return vaddv_u8(x0);
#else
    uint8x8_t x1 = vpadd_u8(x0, x0);
    uint8x8_t x2 = vpadd_u8(x1, x1);
    uint8x8_t x3 = vpadd_u8(x2, x2);
    return vget_lane_u8(x3, 0);
#endif
}
#else
extern NLIB_VIS_PUBLIC const unsigned char _nlib_popcnt_array[];
static NLIB_ALWAYS_INLINE int nlib_popcnt32(uint32_t x) {
    return _nlib_popcnt_array[(x >> 24) & 0xFF] + _nlib_popcnt_array[(x >> 16) & 0xFF] +
           _nlib_popcnt_array[(x >> 8) & 0xFF] + _nlib_popcnt_array[(x)& 0xFF];
}
static NLIB_ALWAYS_INLINE int nlib_popcnt64(uint64_t x) {
    return _nlib_popcnt_array[(x >> 56) & 0xFF] + _nlib_popcnt_array[(x >> 48) & 0xFF] +
           _nlib_popcnt_array[(x >> 40) & 0xFF] + _nlib_popcnt_array[(x >> 32) & 0xFF] +
           _nlib_popcnt_array[(x >> 24) & 0xFF] + _nlib_popcnt_array[(x >> 16) & 0xFF] +
           _nlib_popcnt_array[(x >> 8) & 0xFF] + _nlib_popcnt_array[(x)& 0xFF];
}
static NLIB_ALWAYS_INLINE int nlib_popcnt16(uint16_t x) {
    return _nlib_popcnt_array[(x >> 8) & 0xFF] + _nlib_popcnt_array[(x)& 0xFF];
}
#endif
#define nlib_popcnt(x) nlib_popcnt32(x)

// nlib_clz32(0x80000000) -> 0, nlib_clz32(1) -> 31
static int nlib_clz32(uint32_t x) NLIB_ATTRIBUTE_CONST;
// nlib_ctz32(0x80000000) -> 31, nlib_ctz32(1) -> 0
static int nlib_ctz32(uint32_t x) NLIB_ATTRIBUTE_CONST;
// nlib_clz64(INT64_MIN) -> 0, nlib_clz64(1) -> 63
static int nlib_clz64(uint64_t x) NLIB_ATTRIBUTE_CONST;
// nlib_ctz64(INT64_MIN) -> 63, nlib_ctz64(1) -> 0
static int nlib_ctz64(uint64_t x) NLIB_ATTRIBUTE_CONST;
#if defined(_MSC_VER)
static NLIB_ALWAYS_INLINE int nlib_clz32(uint32_t x) {
    DWORD cnt;
    return _BitScanReverse(&cnt, x) ? (int)(31 - cnt) : 32;  // NOLINT
    // return (int)(__lzcnt(x));  // needs haswell+
}
static NLIB_ALWAYS_INLINE int nlib_ctz32(uint32_t x) {
    DWORD cnt;
    return _BitScanForward(&cnt, x) ? cnt : 32;
}
static NLIB_ALWAYS_INLINE int nlib_clz64(uint64_t x) {
#ifdef NLIB_64BIT
    DWORD cnt;
    return _BitScanReverse64(&cnt, x) ? (int)(63 - cnt) : 64;  // NOLINT
    // return (int)(__lzcnt64(x));  // needs haswell+
#else
    DWORD cnt;
    DWORD dw = (DWORD)(x >> 32);
    if (_BitScanReverse(&cnt, dw)) {
        return (int)(31 - cnt);  // NOLINT
    } else {
        dw = (DWORD)(x);
        return _BitScanReverse(&cnt, dw) ?
            (int)(63 - cnt) : 64;  // NOLINT
    }
#endif
}
static NLIB_ALWAYS_INLINE int nlib_ctz64(uint64_t x) {
#ifdef NLIB_64BIT
    DWORD cnt;
    return _BitScanForward64(&cnt, x) ? cnt : 64;
#else
    DWORD cnt;
    DWORD dw = (DWORD)(x);
    if (_BitScanForward(&cnt, dw)) {
        return (int)(cnt);  // NOLINT
    } else {
        dw = (DWORD)(x >> 32);
        return _BitScanForward(&cnt, dw) ?
            (int)(32 + cnt) : 64;  // NOLINT
    }
#endif
}
#elif defined(CAFE)
static NLIB_ALWAYS_INLINE int nlib_clz32(uint32_t x) { return __CLZ32(x); }
static NLIB_ALWAYS_INLINE int nlib_ctz32(uint32_t x) { return 32 - nlib_clz32(~x & (x - 1)); }
static NLIB_ALWAYS_INLINE int nlib_clz64(uint64_t x) {
    int cnt;
    unsigned int dw = (unsigned int)(x >> 32);  // NOLINT
    cnt = __CLZ32(dw);
    if (cnt < 32) {
        return cnt;
    } else {
        dw = (unsigned int)(x);  // NOLINT
        cnt = __CLZ32(dw);
        return 32 + cnt;
    }
}
static NLIB_ALWAYS_INLINE int nlib_ctz64(uint64_t x) { return 64 - nlib_clz64(~x & (x - 1)); }
#elif defined(NN_PLATFORM_CTR)
static NLIB_ALWAYS_INLINE int nlib_clz32(uint32_t x) { return x != 0 ? __builtin_clz(x) : 32; }
static NLIB_ALWAYS_INLINE int nlib_ctz32(uint32_t x) { return 32 - nlib_clz32(~x & (x - 1)); }
static NLIB_ALWAYS_INLINE int nlib_clz64(uint64_t x) { return x != 0 ? __builtin_clzll(x) : 64; }
static NLIB_ALWAYS_INLINE int nlib_ctz64(uint64_t x) { return 64 - nlib_clz64(~x & (x - 1)); }
#else
static NLIB_ALWAYS_INLINE int nlib_clz32(uint32_t x) { return x != 0 ? __builtin_clz(x) : 32; }
static NLIB_ALWAYS_INLINE int nlib_ctz32(uint32_t x) { return x != 0 ? __builtin_ctz(x) : 32; }
static NLIB_ALWAYS_INLINE int nlib_clz64(uint64_t x) { return x != 0 ? __builtin_clzll(x) : 64; }
static NLIB_ALWAYS_INLINE int nlib_ctz64(uint64_t x) { return x != 0 ? __builtin_ctzll(x) : 64; }
#endif
#define nlib_clz(x) nlib_clz32(x)
#define nlib_ctz(x) nlib_ctz32(x)

static size_t nlib_strlcpy(char* __restrict s1, const char* __restrict s2, size_t s1max)
    NLIB_NONNULL;
static NLIB_C_INLINE
size_t nlib_strlcpy(char* __restrict s1, const char* __restrict s2, size_t s1max) {
#if defined(__FreeBSD__)
    return strlcpy(s1, s2, s1max);
#else
    size_t len = nlib_strlen(s2);
    if (NLIB_LIKELY(len < s1max)) {
        NLIB_MEMCPY(s1, s2, len + 1);
    } else if (NLIB_LIKELY(s1max > 0)) {
        NLIB_MEMCPY(s1, s2, s1max - 1);
        s1[s1max - 1] = '\0';
    }
    return len;
#endif
}

static uint32_t nlib_bitreverse32(uint32_t x) NLIB_ATTRIBUTE_CONST;
static uint64_t nlib_bitreverse64(uint64_t x) NLIB_ATTRIBUTE_CONST;

static NLIB_ALWAYS_INLINE uint32_t nlib_bitreverse32(uint32_t x) {
#if __has_builtin(__builtin_bitreverse32)
    return __builtin_bitreverse32(x);
#elif __has_builtin(__builtin_arm_rbit)
    return __builtin_arm_rbit(x);
#elif defined(__arm__) && !defined(NN_PLATFORM_CTR)
    return __rbit(x);
#else
    x = ((x & 0x55555555UL) << 1) | ((x >> 1) & 0x55555555UL);
    x = ((x & 0x33333333UL) << 2) | ((x >> 2) & 0x33333333UL);
    x = ((x & 0x0F0F0F0FUL) << 4) | ((x >> 4) & 0x0F0F0F0FUL);
#ifdef _MSC_VER
    x = _byteswap_ulong(x);
#elif defined(CAFE) || defined(NN_PLATFORM_CTR)
    x = (x << 24) | ((x & 0xFF00) << 8) |
        ((x >> 8) & 0xFF00) | (x >> 24);
#else
    x = __builtin_bswap32(x);
#endif
    return x;
#endif
}

static NLIB_ALWAYS_INLINE uint64_t nlib_bitreverse64(uint64_t x) {
#if __has_builtin(__builtin_bitreverse64)
    return __builtin_bitreverse64(x);
#elif __has_builtin(__builtin_arm_rbit64)
    return __builtin_arm_rbit64(x);
#elif __has_builtin(__builtin_arm_rbit)
    return __builtin_arm_rbit(x >> 32) |
           (((uint64_t)__builtin_arm_rbit(x)) << 32);
#elif defined(__arm__) && !defined(NN_PLATFORM_CTR)
    return __rbit(x >> 32) |
        (((uint64_t)__rbit(x)) << 32);
#else
    x = ((x & 0x5555555555555555ULL) << 1) | ((x >> 1) & 0x5555555555555555ULL);
    x = ((x & 0x3333333333333333ULL) << 2) | ((x >> 2) & 0x3333333333333333ULL);
    x = ((x & 0x0F0F0F0F0F0F0F0FULL) << 4) | ((x >> 4) & 0x0F0F0F0F0F0F0F0FULL);
#ifdef _MSC_VER
    x = _byteswap_uint64(x);
#elif defined(CAFE) || defined(NN_PLATFORM_CTR)
    x =
        (x << 56) |
        ((x & 0xFF00U) << 40) |
        ((x & 0xFF0000U) << 24) |
        ((x & 0xFF000000U) << 8) |
        ((x >> 8) & 0xFF000000U) |
        ((x >> 24) & 0xFF0000U) |
        ((x >> 40) & 0xFF00U) |
        (x >> 56);
#else
    x = __builtin_bswap64(x);
#endif
    return x;
#endif
}

#undef NLIB_MEMCPY
#undef NLIB_MEMMOVE
#undef NLIB_MEMSET

#ifdef __cplusplus
}
#endif

#if defined(_MSC_VER)
#if defined(n_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#elif defined(nx_misc_EXPORTS)
# undef NLIB_VIS_PUBLIC
# define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif
#endif

#endif  // INCLUDE_NN_NLIB_PLATFORM_H_
