﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SINGLETON_H_
#define INCLUDE_NN_NLIB_SINGLETON_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

template <class T>
class Singleton NLIB_FINAL {
 public:
    static T& Instance() {
        if (!flag_) {
            if (!nlib_atomic_load32(&flag_, NLIB_ATOMIC_ACQUIRE))
                nlib_once(&once_flag_, Init);
        }
        T* ptr = reinterpret_cast<T*>(&memory_[0]);
        return *ptr;
    }
    static T& GetInstance() { return Instance(); }

 private:
    static void Init();
    static int32_t flag_;
    static nlib_onceflag once_flag_;
    // Memory leak would be mis-detected if dynamically allocated.
    NLIB_ALIGNAS(8) static unsigned char memory_[sizeof(T)];

 private:
    Singleton() {}
    NLIB_DISALLOW_COPY_AND_ASSIGN(Singleton);
};

template <class T>
void Singleton<T>::Init() {
    new (memory_) T();
    nlib_atomic_store32(&flag_, 1, NLIB_ATOMIC_RELEASE);
}

template <class T>
nlib_onceflag Singleton<T>::once_flag_ = NLIB_ONCE_INIT;
template <class T>
int32_t Singleton<T>::flag_ = 0;
template <class T>

#ifndef NLIB_DOXYGEN
#if !defined(NN_PLATFORM_CTR)
NLIB_ALIGNAS(8) unsigned char Singleton<T>::memory_[sizeof(T)];
#else
unsigned char Singleton<T>::memory_[sizeof(T)];
#endif
#endif

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SINGLETON_H_
