﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_ZLIBOUTPUTSTREAM_H_
#define INCLUDE_NN_NLIB_ZLIBOUTPUTSTREAM_H_

#include "nn/nlib/Platform.h"
#ifdef NLIB_HAS_ZLIB

#include "nn/nlib/OutputConverterStream.h"

NLIB_NAMESPACE_BEGIN

class OutputStream;

struct ZlibOutputStreamSettings {
    typedef void* (*ReallocFunc)(void* p, size_t n);
    void* buf_in;
    void* buf_out;
    size_t buffer_size;
    int level;
    ReallocFunc realloc_func;

 public:
    NLIB_CEXPR ZlibOutputStreamSettings() NLIB_NOEXCEPT : buf_in(NULL),
                                                          buf_out(NULL),
                                                          buffer_size(64 * 1024),
                                                          level(-1),
                                                          realloc_func(::nlib_realloc) {}
    NLIB_CEXPR ZlibOutputStreamSettings(void* buf_in_, void* buf_out_, size_t buffer_size_,
                                        int level_) NLIB_NOEXCEPT : buf_in(buf_in_),
                                                                    buf_out(buf_out_),
                                                                    buffer_size(buffer_size_),
                                                                    level(level_),
                                                                    realloc_func(::nlib_realloc) {}
    NLIB_CEXPR ZlibOutputStreamSettings(void* buf_in_, void* buf_out_, size_t buffer_size_,
                                        int level_, ReallocFunc realloc_func_) NLIB_NOEXCEPT
        : buf_in(buf_in_),
          buf_out(buf_out_),
          buffer_size(buffer_size_),
          level(level_),
          realloc_func(realloc_func_) {}
};

class NLIB_VIS_PUBLIC ZlibOutputStream NLIB_FINAL : public OutputConverterStream {
 public:
    ZlibOutputStream() NLIB_NOEXCEPT : prv_(NULL) {}
    virtual ~ZlibOutputStream() NLIB_NOEXCEPT NLIB_OVERRIDE;
    errno_t Init() NLIB_NOEXCEPT {
        NLIB_CEXPR ZlibOutputStreamSettings settings;
        return this->Init(settings);
    }
    errno_t Init(const ZlibOutputStreamSettings& settings) NLIB_NOEXCEPT;

 private:
    struct ZlibOutputStreamPrivate;
    ZlibOutputStreamPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ZlibOutputStream);
};

NLIB_NAMESPACE_END

#endif
#endif  // INCLUDE_NN_NLIB_ZLIBOUTPUTSTREAM_H_
