﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SIMD_SIMDQUATERNION_H_
#define INCLUDE_NN_NLIB_SIMD_SIMDQUATERNION_H_

#include "nn/nlib/simd/SimdFloat.h"
#include "nn/nlib/simd/SimdVector4.h"

NLIB_NAMESPACE_BEGIN
namespace simd {

class NLIB_VIS_HIDDEN Quaternion {
 public:
    static SimdQuaternion __vectorcall Identity() NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall Conjugate(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static f128 __vectorcall Length(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static f128 __vectorcall LengthSq(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static f128 __vectorcall RecpLength(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall Normalize(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall NormalizeEst(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall Inverse(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall Ln(SimdQuaternionArg q_normalized) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall Exp(SimdQuaternionArg q) NLIB_NOEXCEPT;

    static bool __vectorcall IsIdentity(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static bool __vectorcall IsInfinite(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static bool __vectorcall IsNaN(SimdQuaternionArg q) NLIB_NOEXCEPT;
    static bool __vectorcall CmpEq(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT;
    static bool __vectorcall CmpNe(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT;

    static f128 __vectorcall Dot(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall
        Mult(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT;

    static SimdQuaternion __vectorcall
        FromRotationAxisAndSinCos(SimdVectorArg axis_normalized, float sin_half_rad,
                                  float cos_half_rad) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall FromRotationMatrix(SimdMatrixArg m) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall
        FromRotationZXY(SimdVectorArg sin_half_xyz, SimdVectorArg cos_half_xyz) NLIB_NOEXCEPT;
    static SimdVector __vectorcall ToAxisAngle(float* rad, SimdQuaternion q) NLIB_NOEXCEPT;

    static SimdQuaternion __vectorcall Slerp(SimdQuaternionArg q0_normalized,
                                             SimdQuaternionArg q1_normalized,
                                             float t) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall
        Squad(SimdQuaternionArg q0_normalized, SimdQuaternionArg q1_normalized,
              SimdQuaternionArg q2_normalized, SimdQuaternionArg q3_normalized,
              float t) NLIB_NOEXCEPT;
    static void __vectorcall SquadSetup(SimdQuaternion* a, SimdQuaternion* b, SimdQuaternion* c,
                                        SimdQuaternionArg q0, SimdQuaternionArg q1,
                                        SimdQuaternionArg q2, SimdQuaternionArg q3) NLIB_NOEXCEPT;
    static SimdQuaternion __vectorcall BaryCentric(SimdQuaternionArg q0, SimdQuaternionArg q1,
                                                   SimdQuaternionArg q2, float f,
                                                   float g) NLIB_NOEXCEPT;

 private:
    Quaternion();  // forbidden
};

#ifndef NLIB_DOXYGEN

#define NLIB_M(tp) inline tp __vectorcall

// XMQuaternionIdentity
NLIB_M(SimdQuaternion) Quaternion::Identity() NLIB_NOEXCEPT { return F128::Set0001(); }

// XMQuaternionConjugate
NLIB_M(SimdQuaternion) Quaternion::Conjugate(SimdQuaternionArg q) NLIB_NOEXCEPT {
    return F128::NegateEx<true, true, true, false>(q);
}

// XMQuaternionLength
NLIB_M(f128) Quaternion::Length(SimdQuaternionArg q) NLIB_NOEXCEPT { return Vector4::Length(q); }

// XMQuaternionLengthSq
NLIB_M(f128) Quaternion::LengthSq(SimdQuaternionArg q) NLIB_NOEXCEPT {
    return Vector4::LengthSq(q);
}

// XMQuaternionReciprocalLength
NLIB_M(f128) Quaternion::RecpLength(SimdQuaternionArg q) NLIB_NOEXCEPT {
    return Vector4::RecpLength(q);
}

// XMQuaternionNormalize
NLIB_M(SimdQuaternion) Quaternion::Normalize(SimdQuaternionArg q) NLIB_NOEXCEPT {
    return Vector4::Normalize(q);
}

// XMQuaternionNormalizeEst
NLIB_M(SimdQuaternion) Quaternion::NormalizeEst(SimdQuaternionArg q) NLIB_NOEXCEPT {
    return Vector4::NormalizeEst(q);
}

// XMQuaternionInverse
NLIB_M(SimdQuaternion) Quaternion::Inverse(SimdQuaternionArg q) NLIB_NOEXCEPT {
    f128 lenSq = LengthSq(q);
    SimdQuaternion conj = Conjugate(q);

    f128 eps = F128::SetEpsilon();
    f128 nearZero = F128::CmpLe(lenSq, eps);

    SimdQuaternion inv = F128::Div(conj, lenSq);
    return F128::AndNot(nearZero, inv);
}

// XMQuaternionLn
NLIB_M(SimdQuaternion) Quaternion::Ln(SimdQuaternionArg q_normalized) NLIB_NOEXCEPT {
    static const float one_eps = 1.f - 0.00001f;
    f128 q0 = F128::SetZeroToLane<3>(q_normalized);
    f128 w = F128::SetValue<3>(q_normalized, each_select32);
    f128 xyz_len = Vector4::Length(q0);
    f128 w_not_near_one = F128::InBound(w, F128::SetValue(one_eps, each_float));

    f128 theta = F128::ArcTan2(xyz_len, w);
    f128 result = F128::Div(theta, xyz_len);
    result = F128::Mult(q0, result);
    result = F128::Select(w_not_near_one, result, q0);
    return result;
}

// XMQuaternionExp
NLIB_M(SimdQuaternion) Quaternion::Exp(SimdQuaternionArg q) NLIB_NOEXCEPT {
    f128 xyz_len = Vector4::Length(F128::SetZeroToLane<3>(q));
    f128x2 sc = F128::SinCos(xyz_len);

    f128 result = F128::Mult(q, F128::Div(sc.val[0], xyz_len));
    f128 near_zero = F128::CmpNearEqZero(xyz_len, F128::SetEpsilon());
    result = F128::Select(near_zero, q, result);
    result = F128::Splat<false, false, false, true>(result, sc.val[1]);
    return result;
}

// XMQuaternionIsIdentity
NLIB_M(bool) Quaternion::IsIdentity(SimdQuaternionArg q) NLIB_NOEXCEPT {  // NOLINT
    return Vector4::CmpEq(q, Identity());
}

// XMQuaternionIsInfinite
NLIB_M(bool) Quaternion::IsInfinite(SimdQuaternionArg q) NLIB_NOEXCEPT {  // NOLINT
    return Vector4::IsInfinite(q);
}

// XMQuaternionIsNaN
NLIB_M(bool) Quaternion::IsNaN(SimdQuaternionArg q) NLIB_NOEXCEPT {  // NOLINT
    return Vector4::IsNaN(q);
}

// XMQuaternionEqual
NLIB_M(bool) Quaternion::CmpEq(SimdQuaternionArg q0,  // NOLINT
                               SimdQuaternionArg q1) NLIB_NOEXCEPT {
    return Vector4::CmpEq(q0, q1);
}

// XMQuaternionNotEqual
NLIB_M(bool) Quaternion::CmpNe(SimdQuaternionArg q0,  // NOLINT
                               SimdQuaternionArg q1) NLIB_NOEXCEPT {
    return Vector4::CmpNe(q0, q1);
}

// XMQuaternionDot
NLIB_M(f128) Quaternion::Dot(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT {
    return Vector4::Dot(q0, q1);
}

// XMQuaternionMultiply
NLIB_M(SimdQuaternion) Quaternion::Mult(SimdQuaternionArg q0, SimdQuaternionArg q1) NLIB_NOEXCEPT {
    SimdVector v1 = F128::Swizzle<3, 0, 1, 2>(q1);
    f128 r1 = F128::Swizzle<3, 2, 1, 0>(q0);
    f128 r2 = F128::Swizzle<2, 3, 0, 1>(q0);
    f128 r3 = F128::Swizzle<1, 0, 3, 2>(q0);
    SimdMatrix m;
    m.r[0] = q0;
    m.r[1] = F128::NegateEx<false, true, false, true>(r1);
    m.r[2] = F128::NegateEx<false, false, true, true>(r2);
    m.r[3] = F128::NegateEx<true, false, false, true>(r3);
    return Vector4::Transform(v1, m);
}

// sin_half_rad = sin(rad/2), cos_half_rad = cos(rad/2)
// XMQuaternionRotationNormal
NLIB_M(SimdQuaternion) Quaternion::FromRotationAxisAndSinCos(SimdVectorArg axis_normalized,
                                                             float sin_half_rad,
                                                             float cos_half_rad) NLIB_NOEXCEPT {
    SimdVector axis = F128::SetFloatToLane<3>(axis_normalized, 1.f);
    f128 scale = F128::SetValue(sin_half_rad, sin_half_rad, sin_half_rad, cos_half_rad);
    return F128::Mult(axis, scale);
}

// m must be rotation matrix
NLIB_M(SimdQuaternion) Quaternion::FromRotationMatrix(SimdMatrixArg m) NLIB_NOEXCEPT {
    // 1-2y^2-2z^2 2xy+2wz     2xz-2wy
    // 2xy-2wz     1-2x^2-2z^2 2yz+2wx
    // 2xz+2wy     2yz-2wx     1^2x^2^2y^2
    f128 r0 = m.r[0];
    f128 r1 = m.r[1];
    f128 r2 = m.r[2];

    f128 elem;
    // elem = {  m00 - m11 - m22 + 1.f, -m00 + m11 - m22 + 1.f,
    //          -m00 - m11 + m22 + 1.f,  m00 + m11 + m22 + 1.f }
    // -> { 4x^2, 4y^2, 4z^2, 4w^2 }
    {
        f128 m00x = F128::SetValue<0>(r0, each_select32);
        f128 m11x = F128::SetValue<1>(r1, each_select32);
        f128 m22x = F128::SetValue<2>(r2, each_select32);
        m00x = F128::NegateEx<false, true, true, false>(m00x);
        m11x = F128::NegateEx<true, false, true, false>(m11x);
        m22x = F128::NegateEx<true, true, false, false>(m22x);
        f128 one = F128::SetOne();
        elem = F128::Add(m00x, m11x);
        elem = F128::Add(elem, m22x);
        elem = F128::Add(elem, one);
    }

    f128 xx_ge_yy;
    f128 zz_ge_ww;
    f128 xxyy_ge_zzww;
    f128 elem_max;
    {
        f128 t0, t1, t2, t3;
        t0 = F128::SetValue<0>(elem, each_select32);
        t1 = F128::SetValue<1>(elem, each_select32);
        t2 = F128::SetValue<2>(elem, each_select32);
        t3 = F128::SetValue<3>(elem, each_select32);
        xx_ge_yy = F128::CmpGe(t0, t1);
        zz_ge_ww = F128::CmpGe(t2, t3);

        t0 = F128::PairwiseMax(elem, elem);
        t2 = F128::SetValue<0>(t0, each_select32);
        t3 = F128::SetValue<1>(t0, each_select32);
        elem_max = F128::PairwiseMax(t0, t0);
        xxyy_ge_zzww = F128::CmpGe(t2, t3);
    }

    f128 v0_25 = F128::SetValue(0.25f, each_float);
    elem_max = F128::Mult(v0_25, elem_max);
    f128 mult = F128::RecpSqrt(elem_max);
    f128 v = F128::Mult(mult, elem_max);
    mult = F128::Mult(v0_25, mult);

    f128 m01_20_12;
    m01_20_12 = F128::Permute<1, 4, 2, -1>(r0, r2);
    m01_20_12 = F128::Permute<0, 1, 6, -1>(m01_20_12, r1);

    f128 m10_02_21;
    m10_02_21 = F128::Permute<0, 1, 5, -1>(r1, r2);
    m10_02_21 = F128::Permute<0, 6, 2, -1>(m10_02_21, r0);

    f128 ans_x_biggest, ans_y_biggest, ans_z_biggest, ans_w_biggest;
    {
        f128 tmp_x, tmp_y, tmp_z, tmp_w;
        tmp_x = F128::NegateEx<false, false, true, true>(m10_02_21);
        tmp_x = F128::Mult(mult, F128::Add(m01_20_12, tmp_x));

        tmp_y = F128::NegateEx<false, true, false, true>(m10_02_21);
        tmp_y = F128::Mult(mult, F128::Add(m01_20_12, tmp_y));

        tmp_z = F128::NegateEx<true, false, false, true>(m10_02_21);
        tmp_z = F128::Mult(mult, F128::Add(m01_20_12, tmp_z));

        tmp_w = F128::Mult(mult, F128::Sub(m01_20_12, m10_02_21));

        ans_x_biggest = F128::Permute<4, 0, 1, 2>(tmp_x, v);
        ans_y_biggest = F128::Permute<0, 4, 2, 1>(tmp_y, v);
        ans_z_biggest = F128::Permute<1, 2, 4, 0>(tmp_z, v);
        ans_w_biggest = F128::Permute<2, 1, 0, 4>(tmp_w, v);
    }

    f128 ans_xy = F128::Select(xx_ge_yy, ans_x_biggest, ans_y_biggest);
    f128 ans_zw = F128::Select(zz_ge_ww, ans_z_biggest, ans_w_biggest);
    return F128::Select(xxyy_ge_zzww, ans_xy, ans_zw);
}

// sin_half_xyz = { sin(rad_x/2), sin(rad_y/2), sin(rad_z/2), * }
// cos_half_xyz = { cos(rad_x/2), cos(rad_y/2), cos(rad_z/2), * }
// XMQuaternionRotationRollPitchYawFromVector
NLIB_M(SimdQuaternion) Quaternion::FromRotationZXY(SimdVectorArg sin_half_xyz,
                                                   SimdVectorArg cos_half_xyz) NLIB_NOEXCEPT {
    //  CxSySz + SxCyCz
    // -SxCySz + CxSyCz
    // -SxSyCz + CxCySz
    //  SxSySz + CxCyCz
    f128 x1 = F128::Permute<4, 0, 0, 0>(sin_half_xyz, cos_half_xyz);
    f128 y1 = F128::Permute<1, 5, 1, 1>(sin_half_xyz, cos_half_xyz);
    f128 z1 = F128::Permute<2, 2, 6, 2>(sin_half_xyz, cos_half_xyz);
    x1 = F128::NegateEx<false, true, true, false>(x1);
    f128 x0 = F128::Permute<0, 4, 4, 4>(sin_half_xyz, cos_half_xyz);
    f128 y0 = F128::Permute<5, 1, 5, 5>(sin_half_xyz, cos_half_xyz);
    f128 z0 = F128::Permute<6, 6, 2, 6>(sin_half_xyz, cos_half_xyz);

    f128 z0x0y0 = F128::Mult(x0, y0);
    f128 z1x1y1 = F128::Mult(x1, y1);
    z0x0y0 = F128::Mult(z0x0y0, z0);
    return F128::MultAdd(z1x1y1, z1, z0x0y0);
}

// XMQuaternionToAxisAngle
// rad can be NULL
NLIB_M(SimdVector) Quaternion::ToAxisAngle(float* rad, SimdQuaternion q) NLIB_NOEXCEPT {
    if (rad) {
        *rad = 2.f * F128::GetFloatFromLane<3>(F128::ArcCos(q));
    }
    return q;
}

// XMQuaternionSlerp
NLIB_M(SimdVector) Quaternion::Slerp(SimdQuaternionArg q0_normalized,
                                     SimdQuaternionArg q1_normalized, float t) NLIB_NOEXCEPT {
    f128 q0q1 = Dot(q0_normalized, q1_normalized);
    f128 tooNear;
    f128 sp;
    {
        f128 ss = F128::MultSub(q0q1, q0q1, F128::SetValue(1.f, each_float));
        f128 eps = F128::SetEpsilon();
        tooNear = F128::CmpLe(ss, eps);
        sp = F128::RecpSqrt(ss);
    }
    f128 ph = F128::ArcCos(q0q1);
    f128 k = F128::SetValue(1.f - t, t, 0.f, 0.f);
    f128 t0t1 = F128::Mult(sp, F128::Sin(F128::Mult(ph, k)));
    f128 t0 = F128::SetValue<0>(t0t1, each_select32);
    f128 t1 = F128::SetValue<1>(t0t1, each_select32);

    SimdVector ret = F128::Mult(q0_normalized, t0);
    ret = F128::MultAdd(q1_normalized, t1, ret);
    return F128::Select(tooNear, q0_normalized, ret);
}

// XMQuaternionSquad
NLIB_M(SimdVector) Quaternion::Squad(SimdQuaternionArg q0_normalized,
                                     SimdQuaternionArg q1_normalized,
                                     SimdQuaternionArg q2_normalized,
                                     SimdQuaternionArg q3_normalized, float t) NLIB_NOEXCEPT {
    float t2 = (t - t * t) * 2.f;
    SimdQuaternion q03 = Slerp(q0_normalized, q3_normalized, t);
    SimdQuaternion q12 = Slerp(q1_normalized, q2_normalized, t);
    return Slerp(q03, q12, t2);
}

// XMQuaternionBaryCentric
NLIB_M(SimdVector) Quaternion::BaryCentric(SimdQuaternionArg q0, SimdQuaternionArg q1,
                                           SimdQuaternionArg q2, float f, float g) NLIB_NOEXCEPT {
    float fg = f + g;
    if (fg <= 0.00001f && fg >= -0.00001f) return q0;  // to avoid division by zero
    SimdQuaternion q01 = Slerp(q0, q1, fg);
    SimdQuaternion q02 = Slerp(q0, q2, fg);
    return Slerp(q01, q02, g / fg);
}

// XMQuaternionSquadSetup
inline void __vectorcall Quaternion::SquadSetup(SimdQuaternion* a, SimdQuaternion* b,
                                                SimdQuaternion* c,
                                                SimdQuaternionArg q0, SimdQuaternionArg q1,
                                                SimdQuaternionArg q2,
                                                SimdQuaternionArg q3) NLIB_NOEXCEPT {
    SimdQuaternion Q0, Q2, Q3;
    {
        f128 lensq_a01 = Quaternion::LengthSq(F128::Add(q0, q1));
        f128 lensq_s01 = Quaternion::LengthSq(F128::Sub(q0, q1));
        f128 cmp01 = F128::CmpLt(lensq_a01, lensq_s01);
        f128 neg_q0 = F128::Negate(q0);
        Q0 = F128::Select(cmp01, neg_q0, q0);

        f128 lensq_a12 = Quaternion::LengthSq(F128::Add(q1, q2));
        f128 lensq_s12 = Quaternion::LengthSq(F128::Sub(q1, q2));
        f128 cmp12 = F128::CmpLt(lensq_a12, lensq_s12);
        f128 neg_q2 = F128::Negate(q2);
        Q2 = F128::Select(cmp12, neg_q2, q2);

        f128 lensq_a23 = Quaternion::LengthSq(F128::Add(q2, q3));
        f128 lensq_s23 = Quaternion::LengthSq(F128::Sub(q2, q3));
        f128 cmp23 = F128::CmpLt(lensq_a23, lensq_s23);
        f128 neg_q3 = F128::Negate(q3);
        Q3 = F128::Select(cmp23, neg_q3, q3);
    }
    const SimdQuaternion& Q1 = q1;

    SimdQuaternion InvQ1 = Quaternion::Inverse(Q1);
    SimdQuaternion Ln_ExpQ1_Q2 = Quaternion::Ln(Quaternion::Mult(InvQ1, Q2));
    SimdQuaternion Ln_ExpQ1_Q0 = Quaternion::Ln(Quaternion::Mult(InvQ1, Q0));

    SimdQuaternion InvQ2 = Quaternion::Inverse(Q2);
    SimdQuaternion Ln_ExpQ2_Q3 = Quaternion::Ln(Quaternion::Mult(InvQ2, Q3));
    SimdQuaternion Ln_ExpQ2_Q1 = Quaternion::Ln(Quaternion::Mult(InvQ2, Q1));

    f128 v0_25 = F128::SetValue(0.25f, each_float);
    SimdQuaternion A = F128::Mult(v0_25, F128::Add(Ln_ExpQ1_Q2, Ln_ExpQ1_Q0));
    SimdQuaternion B = F128::Mult(v0_25, F128::Add(Ln_ExpQ2_Q3, Ln_ExpQ2_Q1));
    A = Quaternion::Exp(A);
    B = Quaternion::Exp(B);

    *a = Quaternion::Mult(Q1, A);
    *b = Quaternion::Mult(Q2, B);
    *c = Q2;
}

#undef NLIB_M

#endif  // NLIB_DOXYGEN

}  // namespace simd
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SIMD_SIMDQUATERNION_H_
