﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_PARAMTYPE_H_
#define INCLUDE_NN_NLIB_TESTING_PARAMTYPE_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/testing/NTest.h"
#include "nn/nlib/testing/ParamTypeTypes.h"

NLIB_NAMESPACE_BEGIN
namespace testing {

template <template <class Tmpl> class TEST_CASE, class LIST>  // NOLINT
class TypedTestLoop {
 public:
    static bool AddTestInfo(const char* test_case, const char* test_name, int index) {
        // NOTE: Executed before main function.
        // This traverses LIST, and AddTestInfo for each type.
        typedef typename LIST::Head T;
        typedef typename LIST::Tail TL;
        typedef TEST_CASE<T> TestCaseType;

        ::nlib_ns::testing::TestInfo::AddTestInfo<TestCaseType>(test_case, test_name, index,
                                                                 TestCaseType::SetUpTestCase,
                                                                 TestCaseType::TearDownTestCase);
        return TypedTestLoop<TEST_CASE, TL>::AddTestInfo(test_case, test_name, index + 1);
    }
};

template <template <class Tmpl> class TEST_CASE>
class TypedTestLoop<TEST_CASE, Tp0> {
 public:
    static bool AddTestInfo(const char* test_case, const char* test_name, int index) {
        NLIB_UNUSED(test_case);
        NLIB_UNUSED(test_name);
        NLIB_UNUSED(index);
        return true;
    }
};

#define NLIB_TESTING_TYPELIST(test_case) ntest_type_params_##test_case##_

}  // namespace testing
NLIB_NAMESPACE_END

#define TYPED_TEST_CASE(test_case, types) typedef types::type NLIB_TESTING_TYPELIST(test_case)

#define TYPED_TEST(test_case, test_name)                                               \
template<class T> \
class NLIB_TESTING_CLASS_NAME_(test_case, test_name)                                   \
        : public test_case<T> {                                                        \
public:                                                                                \
        typedef test_case<T> TestFixture;                                              \
        typedef T TypeParam;                                                           \
        NLIB_TESTING_CLASS_NAME_(test_case, test_name)() {}                            \
private:                                                                               \
        virtual void TestBody();                                                       \
        NLIB_DISALLOW_COPY_AND_ASSIGN(NLIB_TESTING_CLASS_NAME_(test_case, test_name)); \
};                                                                                     \
bool ntest_typed_test_##test_case##test_name##_ = ::nlib_ns::testing::TypedTestLoop<  \
        NLIB_TESTING_CLASS_NAME_(test_case, test_name),                                \
        NLIB_TESTING_TYPELIST(test_case)>::AddTestInfo(#test_case, #test_name, 1);     \
template<class T> \
void NLIB_TESTING_CLASS_NAME_(test_case, test_name)<T>::TestBody()

#endif  // INCLUDE_NN_NLIB_TESTING_PARAMTYPE_H_
