﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_BARRIER_H_
#define INCLUDE_NN_NLIB_THREADING_BARRIER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

// code snippets:
//   Initialization:
//     Barrier barrier;
//     barrier.Init(N);
//   Thread1 .. ThreadN:
//     # do its own job
//     barrier.Wait();  # wait other threads
//     # do its own job
//     barrier.Wait();  # wait other threads again
class Barrier NLIB_FINAL {
 public:
    Barrier() NLIB_NOEXCEPT : is_initialized_(false) {}
    ~Barrier() NLIB_NOEXCEPT {
        if (is_initialized_) {
            errno_t e = nlib_barrier_destroy(&barrier_);
            NLIB_ASSERT_NOERR(e);
            NLIB_UNUSED(e);
        }
    }
    errno_t Init(unsigned int count) NLIB_NOEXCEPT {
        if (is_initialized_) return EAGAIN;
        errno_t e = nlib_barrier_init(&barrier_, count);
        NLIB_ASSERT_NOERR(e);
        if (e != 0) return e;
        is_initialized_ = true;
        return 0;
    }
    errno_t Wait() NLIB_NOEXCEPT {
        NLIB_ASSERT(is_initialized_);
        errno_t e = nlib_barrier_wait(&barrier_);
        NLIB_ASSERT_NOERR(e);
        return e;
    }

 private:
    nlib_barrier barrier_;
    bool is_initialized_;
};

}  // namespace threading
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_THREADING_BARRIER_H_
