#!/usr/bin/env bash

IFS=\

cd $(dirname $0)

if ! type sed > /dev/null 2>&1; then
  echo Please install sed for cygwin.
  echo setup/setup-cygwin-tools.sh would install sed and other tools.
fi

if ! type awk > /dev/null 2>&1; then
  echo Please install sed for cygwin.
  echo setup/setup-cygwin-tools.sh would install awk and other tools.
fi

#
# Find cmake for Windows
#
path_remove() { export PATH=$(echo -n $PATH | /usr/bin/awk -v RS=: -v ORS=: '$0 != "'$1'"' | /usr/bin/sed 's/:$//'); }
path_remove /usr/bin
path_remove /usr/local/bin

WINCMAKE=$(/usr/bin/which cmake)
if [ $? -ne 0 ]; then
  echo cmake for windows not found, please install cmake from https://cmake.org/
  exit 1;
fi
export PATH=/usr/local/bin:/usr/bin:${PATH}

if [ $# -eq 0 ]; then
  if [ "${VS140COMNTOOLS}" = "" ]; then
  if [ "${VS120COMNTOOLS}" = "" ]; then
  if [ "${VS110COMNTOOLS}" = "" ]; then
    echo "Sorry, Visual Studio 2012/2013/2015 not found"
    exit 1
  else
    GEN="Visual Studio 11 2012"
  fi
  else
    GEN="Visual Studio 12 2013"
  fi
  else
    GEN="Visual Studio 14 2015"
  fi
  amd64=" Win64"
  dll=
else
  target=$1
  case ${target} in
  	"vc110" ) GEN="Visual Studio 11 2012" ;;
  	"vc110amd64" ) GEN="Visual Studio 11 2012" ;;
  	"vc110dllamd64" ) GEN="Visual Studio 11 2012" ;;
  	"vc120" ) GEN="Visual Studio 12 2013" ;;
  	"vc120amd64" ) GEN="Visual Studio 12 2013" ;;
  	"vc120dllamd64" ) GEN="Visual Studio 12 2013" ;;
  	"vc140" ) GEN="Visual Studio 14 2015" ;;
  	"vc140amd64" ) GEN="Visual Studio 14 2015" ;;
  	"vc140dllamd64" ) GEN="Visual Studio 14 2015" ;;
  	"vc150amd64" ) GEN="Visual Studio 15 2017" ;;
  	"vc150dllamd64" ) GEN="Visual Studio 15 2017" ;;
  	* ) echo "A Generator script for Visual Studio"
  		echo "Usage: $0 target(e.g. vc120)"
  	    exit 1 ;;
  esac
  case ${target} in
  	*dll* ) dll="-DNLIB_DLL=1" ;;
  	* ) dll= ;;
  esac
  case ${target} in
  	*amd64 ) amd64=" Win64" ;;
  	* ) amd64= ;;
  esac
fi

echo "Generator is ${GEN}${amd64}"

(mkdir -p ${target}/Release && cd ${target}/Release && "$WINCMAKE" -DCMAKE_BUILD_TYPE=Release ${dll} -G "${GEN}${amd64}" ../../)
(mkdir -p ${target}/Debug && cd ${target}/Debug && "$WINCMAKE" -DCMAKE_BUILD_TYPE=Debug ${dll} -G "${GEN}${amd64}" ../../)

