﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#ifndef SAMPLES_SOURCES_MSGPACK_JSONRPC_JSONRPC_H_
#define SAMPLES_SOURCES_MSGPACK_JSONRPC_JSONRPC_H_

#include <stdlib.h>
#include <list>
#include <utility>
#include "nn/nlib/ReallocOutputStream.h"
#include "nn/nlib/threading/CriticalSection.h"

#ifndef NLIB_SOCKET_ENABLED
#define DONT_USE_SOCKET
#endif

class Tunnel {
 public:
    ~Tunnel() {
        using nlib_ns::threading::SimpleCriticalSection;
        using nlib_ns::threading::ScopedLock;
        ScopedLock<SimpleCriticalSection> l(lock_);
        BufType::iterator it = buf_.begin();
        BufType::iterator end = buf_.end();
        for (; it != end; ++it) {
            nlib_free(it->first);
        }
    }
    void Send(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t n) {
        using nlib_ns::threading::SimpleCriticalSection;
        using nlib_ns::threading::ScopedLock;
        ScopedLock<SimpleCriticalSection> l(lock_);
        buf_.push_back(BufItem(data->release(), n));
    }
    bool Receive(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t* n) {
        using nlib_ns::threading::SimpleCriticalSection;
        using nlib_ns::threading::ScopedLock;
        ScopedLock<SimpleCriticalSection> l(lock_);
        if (buf_.empty()) return false;
        BufItem& item = buf_.front();
        data->reset(item.first);
        *n = item.second;
        buf_.pop_front();
        return true;
    }

 private:
    typedef std::pair<uint8_t*, size_t> BufItem;
    typedef std::list<BufItem> BufType;
    nlib_ns::threading::SimpleCriticalSection lock_;
    BufType buf_;
};

extern Tunnel g_c2s;
extern Tunnel g_s2c;

errno_t InitServer();
void ServerThread();
extern volatile bool g_ShutdownServer;

#endif  // SAMPLES_SOURCES_MSGPACK_JSONRPC_JSONRPC_H_
